/*

Copyright (c) <2008, 2009>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <iostream>
#include <time.h>
#include <stdio.h>

#include "charbuf.hpp"

using namespace std;

class NativeTime
{

public:
  int year, month, day;
  int hour, minute, second;
  long sinceEpoch;  // Time since the last epoch

  NativeTime(int yy, int mm, int dd) 
  {
    year = yy;
    month = mm;
    day = dd;
    hour = minute = second = 0;
    calcSinceEpoch();
  }

  NativeTime(int yy, int mm, int dd, int h, int m, int s)
  {
    year = yy;
    month = mm;
    day = dd;
    hour = h;
    minute = m;
    second = s;
    calcSinceEpoch();
  }

  // time is absolute time from Epoch. If timeIsDay = true, then it is
  // measured in days, else in seconds
  NativeTime (long time, bool timeIsDay = true) 
  {
    if (timeIsDay) {
      time = time * 24 * 3600;
    }
    setFromEpoch(time);
  }

  long getSinceEpoch() 
  {
    return sinceEpoch;
  }

  // Caculate, store and return the time since the last epoch
  long calcSinceEpoch() 
  {
    struct tm mytime;
    mytime.tm_sec = second;
    mytime.tm_min = minute;
    mytime.tm_hour = hour;
    mytime.tm_mday = day;
    mytime.tm_mon = month - 1;
    mytime.tm_year = year - 1900;
    mytime.tm_isdst = -1;

    time_t t = mktime (&mytime);
    sinceEpoch = ((long) t);
    return sinceEpoch;
  }
  
  void setFromEpoch(long t)
  {
    struct tm mytime;

    localtime_r((const time_t *) &t, &mytime);

    year = 1900 + mytime.tm_year;
    month = mytime.tm_mon + 1; // tm_mon is in the range 0 to 11
    day = mytime.tm_mday;
    hour = mytime.tm_hour;
    minute = mytime.tm_min;
    second = mytime.tm_sec;

    calcSinceEpoch();
  }

  ostream & print (ostream & out) {
    char s[256];
    toString(s);
    out << s;
    return out;
  }

  charstream & print (charstream & out) {
    out << year << ':';
    out << month << ':';
    out << day << ':';
    out << hour << ':';
    out << minute << ':';
    out << second;
    return out;
  }

  void toString (char * out)
  {
    sprintf(out, "%04d:%02d:%02d:%02d:%02d:%02d", 
	    year, month, day,
	    hour, minute, second);
  }

};

