/*
Copyright (c) <2010>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef __BOGEHOLZ_CONSTS_HPP
#define __BOGEHOLZ_CONSTS_HPP


//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////

// How long (in days) does the simulation run? (Simulation starts at
// time 0)
static const int maxSimTime = 365 * 30; // 30 years

// How many records are active for each resident?
static const int nActiveRecords = 10;

// How many fixed people exist in the facility?
static const int nResident = 30;
static const int nHcw = 2;
static const int nDoctor = 2;
static const int nAdminStaff = 1;
static const int nNurse = 1;
static const int nVolunteer = 2;

// Intervals of fixed periodic events in days
static const int periodDoctorVisit = 7;
static const int periodNurseVisit = 3;
static const int periodHcwVisit = 1;

// Mean spans of various events in days
static const int meanManagerLife = 365 * 5;
static const int meanResidentLife = 365 * 10; // Mean number of days a resident stays in facility
static const int meanHcwLife = 365 * 1;
static const int meanDoctorLife = 365 * 2;
static const int meanAdminStaffLife = 365 * 3;
static const int meanNurseLife = 365 * 2;
static const int meanVolunteerLife = 365 * 1;

static const int meanIncapabilityLength = 15; // Mean number days that a patient incapability lasts for
static const int meanIncapabilityGap = 30 * 4; // Mean number of days between two incapabilities for a single patient

static const int meanGapResidentRead = 30 * 1; // Mean number of days between resident's request to read its records
static const int meanGapExResidentRequest = 30 * 5; // Mean number of days between two requests from ex-residents to obtain their records
static const int meanGapResidentSwitchDoctor = 365 * 1; // Mean number of days before a resident will switch doctors
static const int meanGapManagerResponse = 13; // 13 ~= 30/(ln 10). This ensures that 90% of requests are filled in 30 days
static const int meanGapVolunteerAdd = 2; // Mean number of days between volunteer's addition to progress notes
static const int meanGapCareplanAdd = 30 * 1; // Mean number of days between revisions of careplans

static const int meanGapManagerCheat = 30 * 12; // Mean number of days between two attempts by manager to read random record
static const int meanGapResidentCheat = 30 * 1; // Residents are very curious about others!
static const int meanGapHcwCheat = 30 * 3; 
static const int meanGapAdminStaffCheat = 30 * 3; 
static const int meanGapNurseCheat = 30 * 3; 
static const int meanGapVolunteerCheat = 30 * 12; // Volunteers have little tendency to cheat

static const int meanGapAbuseDisclosure = 30 * 12; // Mean gap between two disclosures of abuse

// Probabilities of various sub-events

static const double doctorReadsInVisit = 0.9; 
// With this probability, doctor will look up 1 in the past
// nActiveRecords dynamic records of each patient in each visit

static const double doctorAddsInVisit = 0.3;
// With this probability, doctor will add a new medical record of each
// patient in each visit

static const double doctorAddsPnInVisit = 0.1;
// With this probability, doctor will add a new private note of each
// patient in each visit

static const double doctorAddsPsychoInVisit = 0.2;
// With this probability, doctor will add a psychotherapy note of each
// patient in each visit

static const double doctorAllowsPsychoAccess = 0.5;
// With this probability, doctor will allow a patient to see
// psychotherapy notes added in the visit

static const double doctorRecommendsPhysioInVisit = 0.1;
// With this probability, the doctor will recommend a physiotherapist
// to each patient in each visit (we assume that the physiotherapist
// so prescribed arrives immediately, so this sub-event immediately
// triggers a physiotherapist visit).

static const double doctorCheatsInVisit = 0.01;
// With this probability, the doctor will read a random record during
// the visit.


static const double physioReadsInVisit = 0.9; 
// With this probability, physiotherapist will look up 1 in the past
// nActiveRecords dynamic records of each patient in each visit

static const double physioAddsInVisit = 0.3;
// With this probability, physiotherapist will add a new rehab note of
// each patient in each visit

static const double physioAddsPnInVisit = 0.1;
// With this probability, physiotherapist will add a new private note
// of each patient in each visit

static const double physioCheatsInVisit = 0.01;
// With this probability, the physiotherapist will read a random
// record during the visit.



static const double nurseReadsInVisit = 0.9; 
// With this probability, nurse will look up 1 in the past
// nActiveRecords dynamic records of each patient in each visit

static const double nurseAddsInVisit = 0.3;
// With this probability, nurse will add a new progress note of each
// patient in each visit


static const double hcwReadsInVisit = 0.2; 
// With this probability, hcw will look up 1 in the past
// nActiveRecords dynamic records of each patient in each visit

static const double hcwAddsInVisit = 0.1;
// With this probability, hcw will add a new progress note of each
// patient in each visit


static const double prConsents = 0.95;
// With this probability, a PR will consent to the modification of
// careplan when asked by the manger


static const double forgetConf = 0.05;
// With this probability, we will forget to sign the confidentiality
// agreement of a new employee

static const double isLawOfficial = 0.95;
// With this probability, a person to whom the manager discloses abuse
// is actually a law official

#endif
