/*
Copyright (c) <2010>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/


#ifndef __BOUNDEDBUFFER_HPP
#define __BOUNDEDBUFFER_HPP

/* A buffer with a fixed length. New entries overwrite oldest
ones. Note that the buffer does not free memory of elements it
overwrites and should, therefore, be used to store value types only,
not pointers (unless an external mechanism tracks and frees the
pointers stored) */

template<class T>
class boundedbuffer
{
protected:
  T * array; // The buffer
  int nextp; // This always points to position where we must write the
	     // next entry.
  int maxLen;
  int currLen;

  // Invariant: if currLen < maxLen, then the data is in positions 0,
  // ..., currLen - 1 and nextp == currLen.

public:
  boundedbuffer(int M) {
    maxLen = M;
    currLen = 0;
    array = new T[maxLen];
    nextp = 0;
  }

  void reset() {
    currLen = nextp = 0;
  }

  void push(T v) {
    array[nextp] = v;
    nextp = (nextp + 1) % maxLen;
    if (currLen < maxLen) currLen ++;
  }

  int length() {
    return currLen;
  }

  // This operation is potentially unsafe because it may return an
  // entry that is not filled or if n >= maxLen large, it may return
  // an entry beyond the end of the array
  T & operator[] (int n) {
    return (array[n]);
  }

  ~boundedbuffer() {
    delete [] array;
  }
};

#endif
