/*
Copyright (c) <2010>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef __DED_SIM_HPP
#define __DED_SIM_HPP

#include <iostream>
#include <queue>

using namespace std;

///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////


class AbstractEvent 
{
protected:
  AbstractEvent() {} // Abstract class

public:
  int time; // Time at which event will trigger
  int priority; // Of all events at time, the priority of this event
  // The actual priority of the event is the lexicographic order
  // (time, priority)

  virtual void fire() = 0; // Whatever the event does goes here

  virtual void printName() {
    cerr << "AbstractEvent" ;
  }

  virtual ~AbstractEvent() {}
};


class evComparison{

public:
  evComparison() {}

  bool operator() (const AbstractEvent * e1, const AbstractEvent * e2) const
  {
    return ((e1->time > e2->time) || 
	    ((e1->time == e2->time) && (e1->priority > e2->priority)));
  }
};

///////////////////////////////////////////////////////////

// Uniform integer in the range [min,max]
int uniform_int (int min, int max); 

// Uniform double in the range [0,1)
double uniform_double ();

// Integer sampled from exponential with mean M
int exp_int (int M);

extern priority_queue<AbstractEvent *, vector<AbstractEvent *>, evComparison> evQ; 
// The main queue of events. The queue is ordered by the < operator
// defined above. evQ.top() returns the event least in the order
// (time, priority).


inline void addEvent(AbstractEvent * ev)
{
  evQ.push(ev);
}

void init_sim(); // User should define this function
void finalize_sim(); // User should define this function

void run_sim(int maxSimTime); // The main function that runs the simulation


#endif
