%% Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
%% All rights reserved.
%%
%% Use in source and binary forms, with or without modification, are
%% permitted provided that the following conditions are met:
%%
%% - Code derived from this source code must retain the above copyright
%%   notice, this list of conditions and the following disclaimer.
%%
%% - Use or modification is not done for any commercial purpose or for
%%   monetary gain
%%
%% Redistribution without permission of original copyright holder is
%% prohibited.
%%
%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
%% "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
%% LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
%% A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
%% HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
%% SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
%% LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
%% DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
%% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
%% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
%% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

%% Sorts 

attr: sort. 
prin: sort. 
info: sort. 
role: sort. 
action: sort. 


%% Principals

str2prin: string -> prin.

%% Info

str2infoid: string -> info.

%% Attributes 

rec: prin -> attr. 
%% rec(k) -- the entire record of principal k

rec/static: prin -> attr. 
%% rec/static(k) -- personal details, financial details of k

rec/dynamic: prin -> date -> attr. 
%% rec/dynamic(k,u) -- careplan, progress notes, etc of k at date u

rec/static/personal: prin -> attr.
rec/static/financial: prin -> attr.
rec/static/emergency: prin -> attr.
rec/static/pastmedical: prin -> attr.

rec/dynamic/careplan: prin -> date -> attr.
rec/dynamic/progressnotes: prin -> date -> attr.
rec/dynamic/medicalrecord: prin -> prin -> date -> attr.

rec/dynamic/medicalrecord/rehab: prin -> prin -> date -> attr.
rec/dynamic/medicalrecord/private: prin -> prin -> date -> attr.
rec/dynamic/medicalrecord/psycho: prin -> prin -> date -> attr.

%%% Roles

resident: role. 
manager: role.
hcw: role.
doctor: prin -> role.
pr: prin -> role.
adminstaff: role.
nurse: role.
volunteer: role.

%%% Actions

add: prin -> info -> action.
%% add k i -- principal k adds information i

delete: prin -> info -> action.
%% delete k i -- principal k deletes information i

read: prin -> info -> action.
%% read k i -- principal k read information i

write: prin -> info -> action.
%% write k i -- principal k writes (updates) information i

disclose: prin -> info -> prin -> action. 
%% disclose k i k' -- principal k discloses information i to principal k'

%%% Dates

%%% Add a relative time (int) to a date
plus: date -> int -> date EVAL.

%%% Integers representing relative time in days
seven_year: int.
one_year: int.
thirty_day: int.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

%%% DB predicates

does: action(-) -> date(-) -> pred DB 
%dbinfo "does" ("action", "time").
%% does a u -- Action a is performed at date u.

request: prin(-) -> date(-) -> pred DB
%dbinfo "request" ("principal", "time").
%% request k u -- Principal k request his info at time u.

beginrole: prin(-) -> role(-) -> date(-) -> pred DB 
%dbinfo "beginrole" ("principal", "role", "time").

endrole: prin(+) -> role(+) -> date(-) -> pred DB 
%dbinfo "endrole" ("principal", "role", "time").

allows: prin(-) -> action(+) -> date(+) -> pred DB 
%dbinfo "allows" ("principal", "action", "time").
%% allows k a u -- k approves of action a at date u.

refer-physio: prin(-) -> prin(+) -> prin(+) -> date(+) -> pred DB 
%dbinfo "refer_physio" ("doctor", "physio", "resident", "time").
%% refer-physio d k k' u -- doctor d refers physiotherapist k to patient k' at date u.

signed-conf: prin(+) -> date(-) -> pred DB
%dbinfo "signed_conf" ("principal", "time").
%% signed-conf k u -- principal k signed a confidentiality agreement at date u.

hasattr: info(+) -> attr(-) -> pred DB
%dbinfo "hasattr" ("info", "attr").
%% hasattr i a -- information i has most specific attribute a
%% This is an internal predicate; the outer predicate is hasattr

%%% EVAL predicates


%% We could also define inrole in terms of beginrole and endrole and
%% just inline it, but a direct implementation might be more
%% efficient.
inrole: prin(+) -> role(+) -> date(+) -> pred EVAL.

%% This is EVAL because of the need to compute the hierarchy on
%% attributes
hasattr_ext: info(+) -> attr(-) -> pred EVAL.
%% hasattr_ext i a -- information i has attribute a

eq_action: action(+) -> action(-) -> pred EVAL.
eq_role: role(+) -> role(-) -> pred EVAL.
eq_prin: prin(+) -> prin(-) -> pred EVAL.
eq_time: date(+) -> date(-) -> pred EVAL.
lessthan: date(+) -> date(+) -> pred EVAL.


%%% SUBJ predicates

%% Used to output violations

violated_action: action(+) -> date(+) -> pred SUBJ.
violated_conf: prin(+) -> role(+) -> date(+) -> pred SUBJ.
violated_request: prin(+) -> date(+) -> pred SUBJ.

emergency: prin(+) -> date(+) -> pred SUBJ.
mentally-incapable: prin(+) -> date(+) -> pred SUBJ.

incorrect: prin(+) -> info(+) -> date(+) -> pred SUBJ. 
%% incorrect k' i u -- record i contains incorrect information about k' at date u.

is-law-official: prin(+) -> date(+) -> pred SUBJ. 
%% Note that law-official is not a role, because it cannot be determined from a trace

believes-victim-of-abuse: prin(+) -> prin(+) -> date(+) -> pred SUBJ.
%% believes-victim-of-abuse k k' u -- k believes at date u that k' is a victim of abuse
