(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

signature DECLPARSER = sig
    val parse_decls: string -> Declaration.decls Option.option
    val parse_decls_file: string -> Declaration.decls Option.option
end


structure DeclParser :> DECLPARSER = struct
structure DeclP = DeclParseFn(DeclLexer)

fun parse_decls (s: string): Declaration.decls Option.option = 
    let val sm = AntlrStreamPos.mkSourcemap()
	val lex = DeclLexer.lex sm
	val strm = DeclLexer.streamify (fn () => "%#parse_decls#% " ^ s ^ " %#parse_end#%")
	val (r, _, errs) = DeclP.parse lex strm
	val errString = String.concatWith "\n" 
			(map (AntlrRepair.repairToString
			      DeclTokens.toString sm) errs)
    in
	(print errString); r
    end
    
(* Read entire file as a string -- be careful, this may cause memory exhaustion *)
fun readFile (f: string): string =
    if (f <> "-") then
	let val inp = BinIO.openIn f
	    val s = BinIO.inputAll inp
	    val s' = Byte.bytesToString s
	in 
	    (BinIO.closeIn(inp); s')
	end
    else (* read stdIn *)
	let fun recread s = 
		let val snew = TextIO.inputLine (TextIO.stdIn)
		in
		    case snew of 
			NONE => s
		      | SOME s' => recread (s ^ s')
		end
	in
	    recread ""
	end

fun parse_decls_file s = parse_decls (readFile s)

end

