(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

%name Decl;

%tokens 
    : ID of string
    | STR of string
    | COLON (":")
    | LIT_SORT ("sort")
    | RARROW ("->")
    | LIT_ANY ("any")
    | LIT_DATE ("date")
    | LIT_INT ("int")
    | LIT_STRING ("string")
    | LIT_SQL ("SQL")
    | LIT_SYM ("SYM")
    | LIT_EVAL ("EVAL")
    | LIT_DB ("DB")
    | LIT_SUBJ ("SUBJ")
    | LIT_INFIX ("%infix")
    | LIT_PRED ("pred")
    | PLUS ("+")
    | MINUS ("-")
    | LPAREN ("(")
    | RPAREN (")")
    | START_DECL
    | DOT (".")
    | COMMA (",")
    | LIT_DBINFO ("%dbinfo")
    ;

%start st;

st
    : START_DECL decls
    ;

decls
    : (decl)*
    ;

decl
    : ID ":" kind "."
        => ( (ID, kind) )
    ;

kind
    : "sort" 
        => ( Declaration.declSort )
    | sortmodelist "pred" predclass inf dbinfo
        => ( Declaration.declPred(sortmodelist, predclass, inf, dbinfo) )
    | sortlist sort constclass@(sort) inf
        => ( Declaration.declConst(sortlist, sort, constclass, inf) )
    ;

sortlist
    : (sort "->")* 
    ;

sortmodelist
    : (sort mode "->" => ( (sort, mode) ))*
    ;

sort
    : "any" => ( Declaration.sortAny )
    | "date" => ( Declaration.sortDate )
    | "int" => ( Declaration.sortInt )
    | "string" => ( Declaration.sortString )
    | ID => ( Declaration.sortOther ID )
    ;

predclass
    : => ( Declaration.predDB )
    | "SQL" => ( Declaration.predSQL )
    | "EVAL" => ( Declaration.predEval )
    | "SUBJ" => ( Declaration.predSubj )
    | "DB" => ( Declaration.predDB )
    ;

constclass(s)
    : => ( case s of
            Declaration.sortDate => Declaration.constEval
          | Declaration.sortString => Declaration.constEval
          | Declaration.sortInt => Declaration.constEval
          | _ => Declaration.constSym
         )
    | "SQL" => ( Declaration.constSQL )
    | "SYM" => ( Declaration.constSym )
    | "EVAL" => ( Declaration.constEval )
    ;

mode
    : ("+" | "(" "+" ")" ) => ( Declaration.modeInput )
    | ("-" | "(" "-" ")" ) => ( Declaration.modeOutput )
    ;

inf
    : => ( Option.NONE )
    | "%infix" STR => ( SOME STR )
    ;

dbinfo
    : => ( NONE )
    | "%dbinfo" STR "(" columnlist ")"
        => ( SOME (STR, columnlist) )
    ;

columnlist
    : STR => ( [STR] )
    | STR "," columnlist 
        => ( STR :: columnlist )
    ;
