(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

%name DeclLexer;

%let endline = [\n\r];
%let ws = [\ \t\n\r];
%let validstrvalue = [^\"\n];

%let stringlimit = "\"";
%let digit = [0-9];
%let lletter = [a-z];
%let uletter = [A-Z];
%let letter = {uletter}|{lletter};
%let special = "_"|"'"|"@"|"-"|"$"|"#"|"/";
%let genchar={lletter}|{uletter}|{digit}|{special};

%let opencomment="(*";
%let closecomment="*)";
%let openlinecomment="%";

%let iden=({letter}|"_")({letter}|{special}|{digit})*;
%let integer=("-")?{digit}+;

%let date=({digit}+":"{digit}+":"{digit}+)(":"{digit}+":"{digit}+":"{digit}+)? ;

%let str="\""{validstrvalue}*"\"";

%let eoflimiter = "%#parse_end#%";

%let start_declarations = "%#parse_decls#%";

%states COMMENT LINECOMMENT;


%defs (
  open DeclTokens
  type lex_result = token
  fun eof() = EOF

  fun extractstring s = String.substring (s, 1, (String.size s) - 2)
);


<INITIAL> {ws} => (continue());
{eoflimiter} => (eof());

<INITIAL> {start_declarations} => ( START_DECL );

<INITIAL> {opencomment} => (YYBEGIN(COMMENT); continue());
<COMMENT> {closecomment} => (YYBEGIN(INITIAL); continue());
<COMMENT> .  => (continue());

<INITIAL> {openlinecomment} => (YYBEGIN(LINECOMMENT); continue());
<LINECOMMENT> {endline} => (YYBEGIN(INITIAL); continue());
<LINECOMMENT> . => (continue());

<INITIAL> ":" => (COLON);
<INITIAL> sort => (LIT_SORT);
<INITIAL> "->" => (RARROW);
<INITIAL> any => (LIT_ANY);
<INITIAL> date => (LIT_DATE);
<INITIAL> int => (LIT_INT);
<INITIAL> string => (LIT_STRING);
<INITIAL> SQL => (LIT_SQL);
<INITIAL> SYM => (LIT_SYM);
<INITIAL> EVAL => (LIT_EVAL);
<INITIAL> DB => (LIT_DB);
<INITIAL> SUBJ => (LIT_SUBJ);
<INITIAL> "%infix" => (LIT_INFIX);
<INITIAL> "%dbinfo" => (LIT_DBINFO);
<INITIAL> pred => (LIT_PRED);
<INITIAL> "+" => (PLUS);
<INITIAL> "-" => (MINUS);
<INITIAL> "(" => (LPAREN);
<INITIAL> ")" => (RPAREN);
<INITIAL> "." => (DOT);
<INITIAL> "," => (COMMA);

<INITIAL> {str} => (STR(extractstring (yytext)));

<INITIAL> {iden} => (ID (yytext));

<INITIAL> . => (print ("Unknown symbol: " ^ yytext ^ "\n"); continue());

