(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

signature FOLPARSER = sig
    val parse_formula: string -> Fol.formula Option.option
    val parse_formula_file: string -> Fol.formula Option.option

    val parse_term: string -> Fol.term Option.option

    (* Parse a term from a string and check that it is a value of a
       given sort in a given signature *)
    val parse_value: Declaration.decls -> Declaration.sort 
		     -> string -> Fol.term Option.option
end


structure FolParser :> FOLPARSER = struct
structure FolP = FolParseFn(FolLexer)

fun parse_formula (s: string): Fol.formula Option.option = 
    let val sm = AntlrStreamPos.mkSourcemap()
	val lex = FolLexer.lex sm
	val strm = FolLexer.streamify (fn () => "%#parse_formula#% " ^ s ^ " %#parse_end#%")
	val (r, _, errs) = FolP.parse lex strm
	val errString = String.concatWith "\n" 
			(map (AntlrRepair.repairToString
			      FolTokens.toString sm) errs)
    in
	(print errString); 
	(Option.map (Fol.preproc_formula o Fol.toFormula) r)
    end
    
fun parse_term (s: string): Fol.term Option.option = 
    let val sm = AntlrStreamPos.mkSourcemap()
	val lex = FolLexer.lex sm
	val strm = FolLexer.streamify (fn () => "%#parse_term#% " ^ s ^ " %#parse_end#%")
	val (r, _, errs) = FolP.parse lex strm
	val errString = String.concatWith "\n" 
			(map (AntlrRepair.repairToString
			      FolTokens.toString sm) errs)
    in
	(print errString); 
	(Option.map Fol.toTerm r)
    end

(* Read entire file as a string -- be careful, this may cause memory exhaustion *)
fun readFile (f: string): string =
    if (f <> "-") then
	let val inp = BinIO.openIn f
	    val s = BinIO.inputAll inp
	    val s' = Byte.bytesToString s
	in 
	    (BinIO.closeIn(inp); s')
	end
    else (* read stdIn *)
	let fun recread s = 
		let val snew = TextIO.inputLine (TextIO.stdIn)
		in
		    case snew of 
			NONE => s
		      | SOME s' => recread (s ^ s')
		end
	in
	    recread ""
	end

fun parse_formula_file s = parse_formula (readFile s)

fun parse_value (ds: Declaration.decls) (s: Declaration.sort) (str: string) =
    case parse_term str of
	NONE => NONE
      | SOME t => (Fol.check_value ds s t;
		   SOME t) handle _ => NONE
end

