(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

%name FolLexer;

%let ws = [\ \t\n\r];
%let validstrvalue = [^\"\n];
%let endline = [\n\r];

%let stringlimit = "\"";
%let digit = [0-9];
%let lletter = [a-z];
%let uletter = [A-Z];
%let letter = {uletter}|{lletter};
%let special = "_"|"'"|"@"|"-"|"$"|"#"|"/";
%let genchar={lletter}|{uletter}|{digit}|{special};

%let opencomment="(*";
%let closecomment="*)";
%let openlinecomment="%";

%let iden=({letter}|"_")({letter}|{special}|{digit})*;
%let integer=(("~") | ("-"))?{digit}+;

%let date=({digit}+":"{digit}+":"{digit}+)(":"{digit}+":"{digit}+":"{digit}+)? ;

%let str="\""{validstrvalue}*"\"";

%let eoflimiter = "%#parse_end#%";

%let start_formula = "%#parse_formula#%";
%let start_term = "%#parse_term#%";


%defs (
  open FolTokens
  type lex_result = token
  fun eof() = EOF

  fun extractstring s = String.substring (s, 1, (String.size s) - 2)


  fun toMonth (1) = Date.Jan
  | toMonth (2) = Date.Feb
  | toMonth (3) = Date.Mar
  | toMonth (4) = Date.Apr
  | toMonth (5) = Date.May
  | toMonth (6) = Date.Jun
  | toMonth (7) = Date.Jul
  | toMonth (8) = Date.Aug
  | toMonth (9) = Date.Sep
  | toMonth (10) = Date.Oct
  | toMonth (11) = Date.Nov
  | toMonth (12) = Date.Dec
  | toMonth n = toMonth (n mod 12 + 1)
  
  fun string2date s = 
  let val sl = String.tokens (fn (#":") => true | _ => false) s
      val ints = map (Option.valOf o Int.fromString) sl
      val ints2 = if (length(ints) = 6) then ints else ints @ [0,0,0]
      val arr = Array.fromList ints2
  in
	 Date.toTime (Date.date ({year = Array.sub(arr,0), month = toMonth(Array.sub(arr,1)),
				  day = Array.sub(arr,2), hour = Array.sub(arr, 3), 
				  minute = Array.sub(arr,4), second = Array.sub(arr, 5),
		                  offset = NONE}))
	(* offset = NONE means that we are measuring local time. This means that to convert 
           back from Time.time to Date.date we must use Date.fromTimeLocal, not 
           Date.fromUTC. 
         *)
  end


);

%states COMMENT LINECOMMENT;


<INITIAL> {ws} => (continue());
{eoflimiter} => (eof());

<INITIAL> {start_formula} => (START_FORMULA);
<INITIAL> {start_term} => (START_TERM);

<INITIAL> {opencomment} => (YYBEGIN(COMMENT); continue());
<COMMENT> {closecomment} => (YYBEGIN(INITIAL); continue());
<COMMENT> .  => (continue());

<INITIAL> {openlinecomment} => (YYBEGIN(LINECOMMENT); continue());
<LINECOMMENT> {endline} => (YYBEGIN(INITIAL); continue());
<LINECOMMENT> . => (continue());

<INITIAL> ":" => (COLON);
<INITIAL> "(" => (LPAREN);
<INITIAL> ")" => (RPAREN);
<INITIAL> "[" => (LSQ);
<INITIAL> "]" => (RSQ);
<INITIAL> "{" => (LBR);
<INITIAL> "}" => (RBR);
<INITIAL> "<" => (LA);
<INITIAL> ">" => (RA);
<INITIAL> all => (LIT_ALL);
<INITIAL> ex => (LIT_EX);
<INITIAL> and => (LIT_AND);
<INITIAL> or => (LIT_OR);
<INITIAL> imp => (LIT_IMP);
<INITIAL> true => (LIT_TRUE);
<INITIAL> false => (LIT_FALSE);
<INITIAL> any => (LIT_ANY);
<INITIAL> date => (LIT_DATE);
<INITIAL> int => (LIT_INT);
<INITIAL> string => (LIT_STRING);
<INITIAL> COMPILE => (LIT_COMPILE);

<INITIAL> {integer} => (INT (Option.valOf (Int.fromString yytext)));
<INITIAL> {date} => (DATE (string2date yytext));

<INITIAL> {str} => (STR(extractstring (yytext)));

<INITIAL> {iden} => (ID (yytext));

<INITIAL> . => (print ("Unknown symbol: " ^ yytext ^ "\n"); continue());

