/*
Copyright (c) <2010>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdlib.h>
#include <math.h>

#include "ded-sim.hpp"


//////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////

// Some basic random number samplers


// Uniform double in the range [0,1)
double uniform_double () 
{
  return (((double) rand())/(((long long) RAND_MAX) + 1));
}

// Uniform integer in the range [min,max)
int uniform_int (int min, int max) 
{
  return ((int) (uniform_double() * (max - min)) + min);
}

// Integer sampled from exponential with mean M
int exp_int (int M)
{
  return (int((-log(uniform_double())) * M));
}



//////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////



priority_queue<AbstractEvent *, vector<AbstractEvent *>, evComparison> evQ;

void run_sim(int maxSimTime)
{
  init_sim();

  while (!evQ.empty()) {

    AbstractEvent * ev = evQ.top();
    evQ.pop();
    
    if (ev -> time >= maxSimTime) 
      // We have exceeded the maximum simulation time; remaining
      // events are useless
      break;

    ev -> fire();

    delete ev;
  }

  // Empty the remaining queue
  while (!evQ.empty()) {
    AbstractEvent * ev = evQ.top();
    evQ.pop();
    delete ev;
  }

  finalize_sim();
}


