/*

Copyright (c) <2010>, <Deepak Garg, garg.deepak@gmail.com>
Copyright (c) <2011>, <Limin Jia, liminjia@cmu.edu>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>


#include <iostream>

#ifdef USE_DB
#include <sqlite3.h>
#endif

#include "charbuf.hpp"
#include "hipaa-consts.hpp"
#include "ded-sim.hpp"
#include "boundedbuffer.hpp"
#include "NativeTime.hpp"
using namespace std;



//////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////

#ifdef USE_DB
sqlite3 * pDb; // Global pointer to database
static const char dbname[]  = "hipaa.db";
#endif

struct rec_t {
  int info_no; 
  int attr_no;
  bool psychoNotes;
};

static const int Hospital = 0; // fixed id for the hospital 

int eventCount = 0;
// A counter to order events at the same time; passed in the priority
// field


int infoCounter = 0;
int msgCounter = 0; // insert eq_message m (info(i,pp))
int purposeCounter = 0;
int attrCounter = 0;
int principalCounter = 0;
int relationCounter = 0;

int patientCounter = 0;
int lawCounter = 0;
int busiCounter = 0;


int currPatients[nPatient];
bool incapable[nPatient];
bool inrelation[nPatient];

boundedbuffer<rec_t> * activeRecords[nPatient];




////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

char * intTimeToString (int t)
{
  static char out[256];
  NativeTime nt(t, true);

  nt.toString (out);
  return out;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////


int run_no_result(const charbuf & cmd_buf)
{
#ifdef USE_DB
  sqlite3_stmt * pStmt;
  
  if (sqlite3_prepare_v2(pDb, cmd_buf.getHead(), -1, &pStmt, NULL) != SQLITE_OK) {
      cerr << "Error preparing the statement:\n<<<<<<<<\n" << cmd_buf << "\n>>>>>>>>\n";
      return 0;
  }
  
  if (sqlite3_step(pStmt) != SQLITE_DONE) {
    cerr << "Error running the statement:\n<<<<<<<<\n" << cmd_buf << "\n>>>>>>>>\n";
    sqlite3_finalize(pStmt);
    return 0;
  }

  sqlite3_finalize(pStmt);
#else
  cout << cmd_buf.getHead() << endl;
#endif
  return 1;
}




///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////

//Patients
class ReplacePatient: public AbstractEvent
// Replace a patient with a new one
{
public:
  int patient_no; // patient number

  ReplacePatient(int t, int p, int p_no)
  {
    time = t;
    priority = p;
    patient_no = p_no;
  }

  void fire();
};

class HospitalVisit: public AbstractEvent
// Patient visit this hospital (cause hospital to disclose for its own TBO)
{
public:
  int patient_no;
  int patient_id;
  
  HospitalVisit(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};


class OtherHospitalVisit: public AbstractEvent
// Patient visit other hospitals (cause disclosure to other covered-entities)
{
public:
  int patient_no;
  int patient_id;
  
  OtherHospitalVisit(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};


class MakePatientIncapable: public AbstractEvent
{
public:
  int patient_no;
  int patient_id;
  
  MakePatientIncapable(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};

class MakePatientCapable: public AbstractEvent
{
public:
  int patient_no;
  int patient_id;
  
  MakePatientCapable(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};


/// / Hospital Employee
// class ReplaceHospitalEmployee: public AbstractEvent
// // employee reports on unethical conduct
// {
// public:
//   int employee_no;
  
//   ReplaceHospitalEmployee(int t, int p, int e_no)
//   {
//     time = t;
//     priority = p;
//     employee_no = e_no;
//   }

//   void fire();
// };

// class EmployeeReportUnethicalPractice: public AbstractEvent
// { // 502/j/1
// public:
//   int employee_no;

//   EmployeeReportUnethicalPractice(int t, int p, int e_no)
//   {
//     time = t;
//     priority = p;
//     employee_no = e_no;
//   }

//   void fire();
  
// };

// class EmployeeReportFindings: public AbstractEvent
// //512b1v
// {
// public:
//   int employee_no;

//   EmployeeReportFindings(int t, int p, int e_no)
//   {
//     time = t;
//     priority = p;
//     employee_no = e_no;
//   }

//   void fire();
// };

// // Business associate

// class ReplaceBusinessAssoc: public AbstractEvent
// // Business associates reports on unethical conduct
// { 
// public:
//   int business_assoc_no;
  
//   ReplaceBusinessAssoc(int t, int p, int b_no)
//   {
//     time = t;
//     priority = p;
//     business_assoc_no = b_no;
//   }

//   void fire();
// };

// class BusiAssocReportUnethicalPractice: public AbstractEvent
// {
//   //502/j/1
// public:
//   int business_assoc_no;

//   BusiAssocReportUnethicalPractice(int t, int p, int b_no)
//   {
//     time = t;
//     priority = p;
//     business_assoc_no = b_no;
//   }

//   void fire();
  
// };


class CourtOrder: public AbstractEvent
{ // 512/e/1/i 
public:

  CourtOrder(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

class LegalProceeding: public AbstractEvent
{ // 512/e/1/ii
public:

  LegalProceeding(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

//Law enforcement



class DiscloseToLawEnforcement: public AbstractEvent
{
  //512/f1

public:
  DiscloseToLawEnforcement(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

class LawEnforcementRequestForIdLoc: public AbstractEvent
{
  //512/f2
public:
  LawEnforcementRequestForIdLoc(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};
class LawEnforcementRequestVictim: public AbstractEvent
{
  // 512/f3/i , f/3/ii
public:
  LawEnforcementRequestVictim(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

// class DiscloseToBusinessAssoc: public AbstractEvent
// {// 502/e/1
// public:

//   DiscloseToBusinessAssoc(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
  
// };

class DiscloseWithConsent: public AbstractEvent
{
  //506/b/1
  public:

  DiscloseWithConsent(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};


class DiscloseForOwnTPO: public AbstractEvent 
//506/c/1
{ 
public: 
  int patient_no;
  int patient_id;
  DiscloseForOwnTPO(int t, int p, int p_no, int p_id){
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};

class DiscloseForOtherTPO: public AbstractEvent 
//506/c/1
{ 
public: 
  int patient_no;
  int patient_id;
  DiscloseForOtherTPO(int t, int p, int p_no, int p_id){
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};

// psychotherapy note is handled by probabilities...

class Marketing: public AbstractEvent
//508
{
public: 
  
  Marketing(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

class DisclosePsychotherapyNotesWithAuth: public AbstractEvent
//508 checked everywhere
{
public:
  rec_t record;
  int patient_id;
  int recv_id;
  int purp_id;

  DisclosePsychotherapyNotesWithAuth(int t, int p, rec_t rec, int p_id, int r_id, int pp_id)
  {
    time = t;
    priority = p;
    record = rec;
    patient_id = p_id;
    recv_id = r_id;
    purp_id = pp_id;
  }

  void fire();
};

class DisclosePsychotherapyNotes: public AbstractEvent
//508 checked everywhere
{
public:
  rec_t record;
  int patient_id;
  int recv_id;
  int purp_id;

  DisclosePsychotherapyNotes(int t, int p, rec_t rec, int p_id, int r_id, int pp_id)
  {
    time = t;
    priority = p;
    record = rec;
    patient_id = p_id;
    recv_id = r_id;
    purp_id = pp_id;
  }

  void fire();
};



class CreateDirectoryInfo: public AbstractEvent
//510/a triggered from hospital visit
{
public: 
  int patient_no;
  int patient_id;
  
  CreateDirectoryInfo(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};

class DiscloseToAssistTPTop: public AbstractEvent
// 510/b/1i triggered from hospital visit
{
public: 
  
  DiscloseToAssistTPTop(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

class DiscloseToAssistTP: public AbstractEvent
// 510/b/1i triggered from hospital visit
{
public: 
  int patient_no;
  int patient_id;
  
  DiscloseToAssistTP(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};

class DiscloseForNotificationTop: public AbstractEvent
// 510/b/1ii triggered from hospital visit 
{
public: 
  DiscloseForNotificationTop(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

class DiscloseForNotification: public AbstractEvent
// 510/b/1ii triggered from hospital visit
{
public: 
  int patient_no;
  int patient_id;
  DiscloseForNotification(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }

  void fire();
};

// class DiscloseForDisasterRelief: public AbstractEvent
// // 510/b/1i
// {
// public: 
  
//   DiscloseForDisasterRelief(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };

class ReportAbuse: public AbstractEvent
//512/c
{
public: 
  
  ReportAbuse(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

// class DiscloseForPublicHealth: public AbstractEvent
// // 512/b/1i
// {
// public: 
  
//   DiscloseForPublicHealth(int t, int p)
//   {
//     time = t;
//     priority = p;
//    }

//   void fire();
// };

// class ReportChildAbuse: public AbstractEvent
// // 512/b/1ii
// {
// public: 
  
//    ReportChildAbuse(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };

// class FDASafety: public AbstractEvent
// // 512/b/1iii
// {
// public: 
  
//   FDASafety(int t, int p)
//   {
//     time = t;
//     priority = p;
//    }

//   void fire();
// };

// class DiscloseForPublicHealthIntervention: public AbstractEvent
// // 512/b/1iv
// {
// public: 
  
//   DiscloseForPublicHealthIntervention(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };

// class DiscloseToHealthOversightAgency: public AbstractEvent
// // 512/d
// {
// public: 
  
//   DiscloseToHealthOversightAgency(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };

class DiscloseToCourt: public AbstractEvent
// 512/e1/i triggered by request
{
public: 
  
  DiscloseToCourt(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};



class DiscloseInLawfulProcessMadeEffortToNotify: public AbstractEvent
// 512/e1/vi triggered by request
{
public: 
  
  DiscloseInLawfulProcessMadeEffortToNotify(int t, int p)
  {
    time = t;
    priority = p;
  }

  void fire();
};

class ReportSuspiciousDeath: public AbstractEvent
// 512/f/4
// can be fired from hospital visit
{
public: 
  int patient_no;
  int patient_id;
  ReportSuspiciousDeath(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
   }

  void fire();
};

class ReportCrimeOnPremise: public AbstractEvent
// 512/f/5
{
public: 
  
  ReportCrimeOnPremise(int t, int p)
  {
    time = t;
    priority = p;
   }

  void fire();
};

class DiscloseToMedicalExaminer: public AbstractEvent
// 512/g/1
// this could be called from hospital visit with some probability
{
public: 
  int patient_no;
  int patient_id;
  DiscloseToMedicalExaminer(int t, int p, int p_no, int p_id)
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
  }
  void fire();
};

class DiscloseToMedicalExaminerBody: public AbstractEvent
// 512/g/1

{
public: 
  int patient_id;
  int receiver_id;
  int purp_id;
  DiscloseToMedicalExaminerBody(int t, int p, int p_id, int r_id, int pp_id)
  {
    time = t;
    priority = p;
    patient_id = p_id;
    receiver_id = r_id;
    purp_id = pp_id;
  }
  void fire();
};

class DiscloseToFuneralDirector: public AbstractEvent
// 512/g/2
// this could be called from hospital visit with some probability
{
public: 
  
  int patient_no;
  int patient_id;
  DiscloseToFuneralDirector(int t, int p, int p_no, int p_id )
  {
    time = t;
    priority = p;
    patient_no = p_no;
    patient_id = p_id;
   }

  void fire();
};

// class DiscloseToOrgan: public AbstractEvent
// // 512/h
// // this could be called from hospital visit with some probability
// {
// public: 
  
//   DiscloseToOrgan(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };

// class DiscloseForResearch: public AbstractEvent
// // 512/i
// {
// public: 
  
//   DiscloseForResearch(int t, int p)
//   {
//     time = t;
//     priority = p;
//    }

//   void fire();
// };

// class DiscloseToLessenRisk: public AbstractEvent
// // 512/j1i
// {
// public: 
  
//   DiscloseToLessenRisk(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };

class DiscloseToIdCriminal: public AbstractEvent
// 512/j1iiA
{
public: 
  
  DiscloseToIdCriminal(int t, int p)
  {
    time = t;
    priority = p;
   }

  void fire();
};

class DiscloseToApprehendEscaped: public AbstractEvent
// 512/j1iiB
{
public: 
  
  DiscloseToApprehendEscaped(int t, int p)
  {
    time = t;
    priority = p;
   }

  void fire();
};

// class DiscloseForWorkerCompensation: public AbstractEvent
// // 512/l
// {
// public: 
  
//   DiscloseForWorkerCompensation(int t, int p)
//   {
//     time = t;
//     priority = p;
//    }

//   void fire();
// };



// class DiscloseWithLimitedDataAgreement: public AbstractEvent
// //514/e
// {
// public: 
  
//   DiscloseWithLimitedDataAgreement(int t, int p)
//   {
//     time = t;
//     priority = p;
//    }

//   void fire();
// };

// class DiscloseForFundRaising: public AbstractEvent
// //514/f
// {
// public: 
  
//   DiscloseForFundRaising(int t, int p)
//   {
//     time = t;
//     priority = p;
//    }

//   void fire();
// };

// class DiscloseInsuranceRelatedInfo: public AbstractEvent
// //514/g
// {
// public: 
  
//   DiscloseInsuranceRelatedInfo(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };




// class ReportCrime: public AbstractEvent
// {
// public:
// ReportCrime(int t, int p)
//   {
//     time = t;
//     priority = p;
//   }

//   void fire();
// };



///////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////

void ReplacePatient::fire ()
{
  //switch out an existing patient

  charbuf stmt(1024);
  int info_id = infoCounter ++;
  int attr_id = attrCounter ++;
  int rel_id = relationCounter ++;
  rec_t rec;

  currPatients[this -> patient_no] = patientCounter++;

  // Reset the active records of patient
  activeRecords[this -> patient_no] -> reset();

  // Add to patient record
  rec.info_no = info_id;
  rec.attr_no = attr_id;
  rec.psychoNotes = false;
  activeRecords[this -> patient_no] -> push(rec);

  // Add attributes
  stmt.reset();
  stmt << "INSERT INTO hasattrof_db VALUES ("
       << "'(str2info \"_i" << info_id << "\")',"
       << "'(str2prin \"_p" << this -> patient_no << "\")'," 
       << "'(str2attr \"_t" << attr_id << "\")', '" 
       << intTimeToString (this -> time) << "');";

  run_no_result(stmt);

  // insert t in phi
  stmt.reset();
  stmt << "INSERT INTO attr_in_db VALUES ("
       << "'(str2attr \"_t" << attr_id << "\")',"
       << "'phi', '"
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);

  // add a relation to the hospital
  inrelation[this -> patient_no] = rel_id;
  stmt.reset();
  stmt << "INSERT INTO inrelation VALUES ("
       << "'(str2prin \"_p" << currPatients[this -> patient_no] << "\")',"
       << "'(str2prin \"_hospital" << Hospital << "\")', " 
       << "'(str2rel \"_r" << rel_id << "\")', ' " 
       << intTimeToString (this -> time) << "');";
  run_no_result(stmt);

  // Initially, patient is capable
  incapable[this -> patient_no] = false;

  // Add incapacitate
  int incaptime = this -> time + exp_int (meanIncapabilityGap);
  addEvent (new MakePatientIncapable(incaptime, eventCount++, this -> patient_no, currPatients[this -> patient_no]));


  // Add hospital visits
  int visittime = this -> time + exp_int (meanGapHospitalVisit);
   addEvent(new HospitalVisit(visittime, eventCount ++, this -> patient_no, 
 			     currPatients[this->patient_no]));

  // Add  visits to other hospitals
  int visittime1 = this -> time + exp_int (meanGapOtherHospitalVisit);
   addEvent(new OtherHospitalVisit(visittime1, eventCount ++, this -> patient_no, 
 			     currPatients[this->patient_no]));

 // Add replace resident
  int replacetime = this -> time + exp_int (meanPatientRet);
  addEvent(new ReplacePatient(replacetime, eventCount ++, this -> patient_no));
  
}

void HospitalVisit::fire()
{
  if (currPatients[this -> patient_no] == this -> patient_id){
  // The patient is still there

    charbuf stmt(1024);
    int info_id = infoCounter ++;
    int attr_id = attrCounter ++;
    rec_t rec;
    
    // Add to patient record
    rec.info_no = info_id;
    rec.attr_no = attr_id;
    rec.psychoNotes = (uniform_double() < createPsychoNotes);
    activeRecords[this -> patient_no] -> push(rec);
    
    // Add attributes
    stmt.reset();
    stmt << "INSERT INTO hasattrof_db VALUES ("
	 << "'(str2info \"_i" << info_id << "\")',"
	 << "'(str2prin \"_p" << this -> patient_no << "\")'," 
	 << "'(str2attr \"_t" << attr_id << "\")', '" 
	 << intTimeToString (this -> time) << "');";

    run_no_result(stmt);

    // insert t in phi
    stmt.reset();
    stmt << "INSERT INTO attr_in_db VALUES ("
	 << "'(str2attr \"_t" << attr_id << "\")',"
	 << "'phi', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);

    if (rec.psychoNotes) {
      // insert t in psychotherapy-notes
      stmt.reset();
      stmt << "INSERT INTO attr_in_db VALUES ("
	   << "'(str2attr \"_t" << attr_id << "\")',"
	   << "'psychotherapy-notes', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    
    // create directory info (510a)
    //    if (uniform_double() < createDirInfo){
    //       addEvent (new CreateDirectoryInfo(this -> time, eventCount ++, this->patient_no, this->patient_id));
    //     };
    
   // disclose for own payment/treatment 506/c/1
    if (uniform_double() < discloseForTPO){
	addEvent (new DiscloseForOwnTPO(this->time, eventCount++, this-> patient_no, this-> patient_id));
      };
 
      // inform family member to assist treatment
    if(uniform_double()<informFamily){
      addEvent(new DiscloseToAssistTP(this->time, eventCount++, this-> patient_no, this-> patient_id));
    };
    
    // Add death to trigger death investigation, funeral director
    if(uniform_double()< deathVisit){
      
      // insert deased p u
      stmt.reset();
      stmt << "INSERT INTO deceased VALUES (" 
	   << "'(str2prin \"_p" <<  this->patient_id << "\")'," << "' "
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);

      if(uniform_double()<suspiciousDeath){
	addEvent (new ReportSuspiciousDeath(this->time, eventCount++, this-> patient_no, this-> patient_id));
	  };
      
      if(uniform_double()<funeral){
	addEvent (new DiscloseToFuneralDirector(this->time, eventCount++, this-> patient_no, this-> patient_id));
	  };
      
      if(uniform_double()<informFamilyDeath){
	addEvent (new DiscloseForNotification(this->time, eventCount++, this-> patient_no, this-> patient_id));
	  };
      
      if(uniform_double()<examineDeath){
	addEvent (new DiscloseToMedicalExaminer(this->time, eventCount++, this-> patient_no, this-> patient_id));
	  };
      
      //replace patient
      addEvent(new ReplacePatient(this->time, eventCount++, this-> patient_no));
    }
    
    int nextVisitTime = this->time + exp_int(meanGapHospitalVisit);
    addEvent(new HospitalVisit(nextVisitTime, eventCount++, this-> patient_no, this->patient_id));
    
  }}

void OtherHospitalVisit::fire()
{

  if ( (currPatients[this -> patient_no] == this -> patient_id)
       && (uniform_double() < discloseDueToOtherVisits)){
    // The patient is still there

    addEvent(new DiscloseForOtherTPO(this->time, eventCount++, this->patient_no, this->patient_id));}

  int nextVisitTime = this -> time + exp_int(meanGapOtherHospitalVisit);
  addEvent(new OtherHospitalVisit(nextVisitTime, eventCount++, this->patient_no, this->patient_id));
}
 

void DiscloseForOtherTPO:: fire (){
  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++; // receipient of the message
    
  //randomly pick a record
  int nRecords = activeRecords[this -> patient_no] -> length();
  int organizedHealthId = principalCounter ++;  

  if (nRecords > 0) {
    charbuf stmt(1024);
    int rec_no = uniform_int(0, nRecords);
    rec_t rec = (*activeRecords[this -> patient_no])[rec_no];
    
    //insert m = msg(info, purpose)
 //    stmt.reset();
//     stmt << "INSERT INTO eq_msg VALUES (" 
// 	 << "'(str2msgid \"_m" <<  msg_id << "\")', "
// 	 << "'(msg (str2info \"_i" << rec.info_id << "\")"
// 	 << "(str2purpid \"_p" << purp_id  << "\"))'"
// 	 << ");";
//     run_no_result(stmt);

    // insert disclose p1 p2 (msg(info, purpose)) u
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << rec.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    // if it is a psychotherapy note then...
    if (rec.psychoNotes){
      addEvent(new DisclosePsychotherapyNotesWithAuth(this->time, eventCount++, rec, this-> patient_id, p2_id, purp_id));
    }

     //506 c2 - c5
    switch(uniform_int(0,4)){
    case 0: //c2
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'covered-entity', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else { // pad the database with an entry
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(health-care-provider (str2prin \"_p" << this-> patient_id <<"\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      if (uniform_double() < purpIn){
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(treatment (str2prin \"_rv" <<  p2_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
      
    case 1: //c3
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'covered-entity', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(health-care-provider (str2prin \"_p" << this-> patient_id <<"\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

      else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      if (uniform_double() < purpIn){
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(payment (str2prin \"_rv" <<  p2_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;

    case 2: {//c4
      int rel_id = relationCounter ++;
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'covered-entity','"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

	if ((uniform_double() < pertainToRel) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_pertain_to VALUES (" 
	     << "'(str2attr \"_t" <<  rec.attr_no << "\")', "
	     << "'(str2rel\"_r" << inrelation[this->patient_id]<< "\")','"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

	if ((uniform_double() < pertainToRel) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_pertain_to VALUES (" 
	     << "'(str2attr \"_t" <<  rec.attr_no << "\")', "
	     << "'(str2rel\"_r" << rel_id << "\")','"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}


      if (uniform_double() < inRel){
	int pasttime = uniform_int(0, this->time);
	stmt.reset();
	stmt << "INSERT INTO inrelation VALUES (" 
	     << "'(str2prin \"_p" <<  this->patient_id << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")', " 
	     << "'(str2rel\"_r" << rel_id << "\")','"
	     << intTimeToString (pasttime) << "');";
	run_no_result(stmt);}
      else {
	int pasttime = uniform_int(0, this->time);
	stmt.reset();
	stmt << "INSERT INTO inrelation VALUES (" 
	     << "'(str2prin \"_p" <<  this->patient_id << "\")',"
	     << "'(str2prin \"_fake" << p2_id << "\")', " 
	     << "'(str2rel\"_r" << rel_id << "\")','"
	     << intTimeToString (pasttime) << "');";
	run_no_result(stmt);}
      
 
      switch(uniform_int(0,3)){
      case 0: 
	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'(healthcare-op-para-1-2 (str2prin \"_rv" 
	       <<  p2_id << "\"))', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      case 1: 
	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'healthcare-fraud-abuse-detection', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      case 2: 
	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'healthcare-fraud-abuse-compliance', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;}
      break;}

    case 3: //c5
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'covered-entity','"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
   
      if (uniform_double() < participantOfOrganizedHealth){

	stmt.reset();
	stmt << "INSERT INTO is_participant_of_organized_health VALUES (" 
	     << "'(str2prin \"_hospital" << Hospital << "\")', "
	     << "'(str2prin \"_o" <<  organizedHealthId << "\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else {
	stmt.reset();
	stmt << "INSERT INTO is_participant_of_organized_health VALUES (" 
	     << "'(str2prin \"_fake" << Hospital << "\")', "
	     << "'(str2prin \"_o" <<  organizedHealthId << "\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

	if ((uniform_double() < belongRole) && genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_belongrole VALUES (" 
	       << "'(str2prin \"_o" <<  organizedHealthId << "\")', "
	       << "'organized-health-arrangement', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	
	if (uniform_double() < participantOfOrganizedHealth){
	  stmt.reset();
	  stmt << "INSERT INTO is_participant_of_organized_health VALUES (" 
	       << "'(str2prin \"_rv" <<  p2_id << "\")', "
	       << "'(str2prin \"_o" <<  organizedHealthId << "\")', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else {
	stmt.reset();
	stmt << "INSERT INTO is_participant_of_organized_health VALUES (" 
	     << "'(str2prin \"_fake" << Hospital << "\")', "
	     << "'(str2prin \"_o" <<  organizedHealthId << "\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'(healthcare-op (str2prin \"_o" 
	    //	       <<  organizedHealthId << "\"))', '"
	       << organizedHealthId << "\"))', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    }}}


void MakePatientIncapable::fire ()
{

if (currPatients[this -> patient_no] == this -> patient_id) {
    // Patient has not left in meantime
    incapable[this -> patient_no] = true;
    
    // Add entry into the subj_begin_incapable table
    charbuf stmt(1024);
    if (genSubjPred){
    stmt << "INSERT INTO subj_begin_incapable VALUES ("
	 << "'(str2prin \"_p" << currPatients[this -> patient_no] << "\")', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);}
    
    // Add an event to make the patientident capable again
    int captime = this -> time + exp_int (meanIncapabilityLength);
    addEvent (new MakePatientCapable(captime, eventCount++, this -> patient_no, currPatients[this -> patient_no]));
  }
  // else ignore the event, because patient has left

}

void MakePatientCapable::fire()
{
  if (currPatients[this -> patient_no] == this -> patient_id) {
    // Patient has not left in meantime
    incapable[this -> patient_no] = false;
    
    // Add entry into the subj_end_incapable table
    charbuf stmt(1024);

    if (genSubjPred){
    stmt << "INSERT INTO subj_end_incapable VALUES ("
	 << "'(str2prin \"_p" << currPatients[this -> patient_no] << "\")', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);}
    
    // Add an event to make the patientident incapable again
    int incaptime = this -> time + exp_int (meanIncapabilityGap);
    addEvent (new MakePatientIncapable(incaptime, eventCount++, this -> patient_no, currPatients[this -> patient_no]));
  }
  // else ignore the event, because patient has left
}

void DisclosePsychotherapyNotesWithAuth::fire()
{
   charbuf stmt(1024);

    if(uniform_double() < obtainedAuthorization){
      int authTime = uniform_int(0, this->time);
      int msgId = msgCounter ++;
      // insert send p p1 m' u'
      stmt.reset();
      stmt << "INSERT INTO send VALUES ("
	   << "'(str2prin \"_p" << patient_id << "\")',"
	   << "'(str2prin \"_hospital" << Hospital << "\")',"
	   << "'(str2msg \"_m" << msgId << "\")','"
	   << intTimeToString (authTime)
	   << "');";
      run_no_result(stmt);

      if((uniform_double() <isValidAuth) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_is_valid_auth VALUES ("
	     << "'(str2msg \"_m" << msgId << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << this-> recv_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << this->record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << this -> purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";}}
    else{
      int authTime = uniform_int(0, this->time);
      int msgId = msgCounter ++;
      // insert send p p1 m' u'
      stmt.reset();
      stmt << "INSERT INTO send VALUES ("
	   << "'(str2prin \"_p" << patient_id << "\")',"
	   << "'(str2prin \"_fake" << Hospital << "\")',"
	   << "'(str2msg \"_m" << msgId << "\")','"
	   << intTimeToString (authTime)
	   << "');";
      run_no_result(stmt);
    }
}

void DisclosePsychotherapyNotes::fire()
{
   charbuf stmt(1024);

   // one of the 3 ways that the disclosure is allowed
  switch (uniform_int(0,3)) {
  case 0:
    if(uniform_double() < obtainedAuthorization){
      int authTime = uniform_int(0, this->time);
      int msgId = msgCounter ++;
      // insert send p p1 m' u'
      stmt.reset();
      stmt << "INSERT INTO send VALUES ("
	   << "'(str2prin \"_p" << patient_id << "\")',"
	   << "'(str2prin \"_hospital" << Hospital << "\")',"
	   << "'(str2msg \"_m" << msgId << "\")','"
	   << intTimeToString (authTime)
	   << "');";
      run_no_result(stmt);

      if(( uniform_double() <isValidAuth) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_is_valid_auth VALUES ("
	     << "'(str2msg \"_m" << msgId << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << this-> recv_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << this->record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << this -> purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";}}
    else{
      int authTime = uniform_int(0, this->time);
      int msgId = msgCounter ++;
      // insert send p p1 m' u'
      stmt.reset();
      stmt << "INSERT INTO send VALUES ("
	   << "'(str2prin \"_p" << patient_id << "\")',"
	   << "'(str2prin \"_fake" << Hospital << "\")',"
	   << "'(str2msg \"_m" << msgId << "\")','"
	   << intTimeToString (authTime)
	   << "');";
      run_no_result(stmt);}
    break; 

  case 1:
    if(uniform_double() < purpIn){
    stmt.reset();
    stmt << "INSERT INTO purp_in_db VALUES (" 
	 << "'(str2purp \"_pp" <<   this -> purp_id << "\")', "
	 << "'(counseling-training-programs (str2prin \"_hospital" 
	 << Hospital << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);
    }
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}; 
    break;

  case 2:
    if (uniform_double() < inRole) {
      stmt.reset();
      stmt << "INSERT INTO inrole_db VALUES (" 
	   << "'(str2prin \"_rv" <<  this-> recv_id << "\")', "
	   << "'secretary','"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
     else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  recv_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
   
    if(uniform_double() < purpIn){
    stmt.reset();
    stmt << "INSERT INTO purp_in_db VALUES (" 
	 << "'(str2purp \"_pp" <<   this -> purp_id << "\")', "
	 << "'(defense-in-legal-proceeding (str2prin \"_hospital"
	 << Hospital << "\") "
	 << "(str2prin \"_p" << this-> patient_id<< "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);};
    break;
  }   
}
	

// void CreateDirectoryInfo::fire()
// { // to do
//  if (currPatients[this -> patient_no] == this -> patient_id){
//   // The patient is still there
   
//  }
// }

void ReportSuspiciousDeath::fire()
{
// 512/f/4

  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int nRecords = activeRecords[this -> patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[this -> patient_no])[nRecords-1];

 //    //insert m =
//     stmt.reset();
//     stmt << "INSERT INTO eq_msg VALUES (" 
// 	 << "'(str2msgid \"_m" <<  msg_id << "\")', "
// 	 << "'(msg (str2info \"_i" << record.info_no << "\")"
// 	 << "(str2purpid \"_p" << purp_id  << "\"))'"
// 	 << ");";
//     run_no_result(stmt);

    // insert send  msg(info, purpose)
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
    
    if ((uniform_double() < belongRole) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_belongrole VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'law-enforcement-official', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
    if ((uniform_double() < believesDeathMaybeResultOfCrime) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_believes_death_may_be_result_of_crime VALUES (" 
	     << "'(str2prin \"_hospital" << Hospital << "\")', "
	     << "'(str2prin \"_p" <<  this->patient_id << "\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'(suspicious-death-notification (str2prin \"_p" 
	   <<  this->patient_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

  }}


void DiscloseToFuneralDirector::fire()
{

  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int nRecords = activeRecords[this -> patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[this -> patient_no])[nRecords-1];

   //insert m = msg(info, purpose)
 //    stmt.reset();
//     stmt << "INSERT INTO eq_msg VALUES (" 
// 	 << "'(str2msgid \"_m" <<  msg_id << "\")', "
// 	 << "'(msg (str2info \"_i" << record.info_no << "\")"
// 	 << "(str2purpid \"_p" << purp_id  << "\"))'"
// 	 << ");";
//     run_no_result(stmt);

    // insert send
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    
    // if it is a psychotherapy note then...
    // authorization
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotes(this->time, eventCount++, record, this-> patient_id, p2_id, purp_id));}


  if (uniform_double() < inRole) {
    stmt.reset();
    stmt << "INSERT INTO inrole_db VALUES (" 
	 << "'(str2prin \"_rv" <<  p2_id << "\")', "
	 << "'funeral-director','"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);}
   else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

  if (uniform_double() < purpIn){
    stmt.reset();
    stmt << "INSERT INTO purp_in_db VALUES (" 
	 << "'(str2purp \"_pp" <<  purp_id << "\")', "
	 << "'(funeral-directory-duties (str2prin \"_p" 
	 <<  this-> patient_id << "\"))', '"
	 << intTimeToString (this -> time) << "');";
    run_no_result(stmt);}
  else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
  
  if ((uniform_double() < necessaryForDuties) && genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_necessary_for_duties VALUES (" 
	   << "'(str2prin \"_hospital" << Hospital << "\")', "
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'(str2prin \"_p" <<  this->patient_id << "\")', "
	   << "'(str2attr \"_t" <<  record.attr_no << "\")', "
	   << "'(str2purp \"_pp" <<  purp_id << "\")', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
  }}

void DiscloseForOwnTPO::fire()
{
      int purp_id = purposeCounter ++;
      int p2_id = principalCounter ++;
      //randomly pick a record
      int nRecords = activeRecords[this -> patient_no] -> length();
      
      if (nRecords > 0) {
	charbuf stmt(1024);
	int rec_no = uniform_int(0, nRecords);
	rec_t rec = (*activeRecords[this -> patient_no])[rec_no];
	
	//insert m = msg(info, purpose)
// 	stmt.reset();
// 	stmt << "INSERT INTO eq_msg VALUES (" 
// 	     << "'(str2msgid \"_m" <<  msg_id << "\")', "
// 	     << "'(msg (str2info \"_i" << rec.info_no << "\")"
// 	     << "(str2purpid \"_p" << purp_id  << "\"))'"
// 	     << ");";
// 	run_no_result(stmt);

	// insert disclose
	stmt.reset();
	stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << rec.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
	run_no_result(stmt);

	// if it is a psychotherapy note then...
	if (rec.psychoNotes){
	  addEvent(new DisclosePsychotherapyNotes(this->time, eventCount++, rec, this-> patient_id, p2_id, purp_id));

	}
	// 506/c/1

	if (uniform_double() < isWorkforceMem){
	  stmt.reset();
	  stmt << "INSERT INTO is_workforce_member VALUES (" 
	       << "'(str2prin \"_rv" <<  p2_id << "\")', "
	       << "'(str2prin \"_hospital" << Hospital << "\")', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else {
	  stmt.reset();
	  stmt << "INSERT INTO is_workforce_member VALUES (" 
	       << "'(str2prin \"_rv" <<  p2_id << "\")', "
	       << "'(str2prin \"_fake" << Hospital << "\")', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}

	if (uniform_double() < inRole){
	  stmt.reset();
	  stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'Health-care-provider', '"
	     << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	 else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

	switch(uniform_int(0, 3)){
	case 0: 
	  if (uniform_double() < purpIn){
	    stmt.reset();
	    stmt << "INSERT INTO purp_in_db VALUES (" 
		 << "'(str2purp \"_pp" <<  purp_id << "\")', "
		 << "'(treatment (str2prin \"_hospital" << Hospital << "\"))', '"
		 << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	  else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	  break;
	  
	case 1: 
	  if (uniform_double() < purpIn){
	    stmt.reset();
	    stmt << "INSERT INTO purp_in_db VALUES (" 
		 << "'(str2purp \"_pp" <<  purp_id << "\")', "
		 << "'(payment (str2prin \"_hospital" << Hospital << "\"))', '"
		 << intTimeToString (this -> time) << "');";
	    run_no_result(stmt);}
	  else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	  break;

	case 2: 
	  if (uniform_double() < purpIn){
	    stmt.reset();
	    stmt << "INSERT INTO purp_in_db VALUES (" 
		 << "'(str2purp \"_pp" <<  purp_id << "\")', "
		 << "'(healthcare-op (str2prin \"_hospital" << Hospital << "\"))', '"
		 << intTimeToString (this -> time) << "');";
	    run_no_result(stmt);}
	  else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	  break;
	}
      }}


void ReportCrimeOnPremise::fire(){ 
// 512/f/5

 int purp_id = purposeCounter ++;
 int p2_id = principalCounter ++;
 int patient_no = uniform_int(0, nPatient);
 int patient_id = currPatients[patient_no];
 int nRecords = activeRecords[patient_no] -> length();
 
  if (nRecords > 0) {
    charbuf stmt(1024);
    int rec_no = uniform_int(0, nRecords);
    rec_t record = (*activeRecords[patient_no])[rec_no];

  // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if ((uniform_double() < belongRole) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_belongrole VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'law-enforcement-official', '"
	     << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}

    if ((uniform_double() < believesEvidenceOfCrimeOnPremises)  && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_believes_evidence_of_crime_on_premises VALUES (" 
	     << "'(str2prin \"_hospital" << Hospital << "\")', "
	     << "'(str2prin \"_p" <<  patient_id << "\")', "
	     << "'(str2attr \"_t" <<  record.attr_no << "\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'(report-possible-crime-on-premises (str2prin \"_hospital" <<  Hospital << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      

  }
 int nextTime = this->time + exp_int(meanGapReportCrime);
  addEvent(new ReportCrimeOnPremise(nextTime, eventCount ++));
}

void DiscloseWithConsent::fire (){
  //506/b/1

  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    int rec_no = uniform_int(0, nRecords);
    rec_t record = (*activeRecords[patient_no])[rec_no];
    
    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
    
    
    // if it is a psychotherapy note then...
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotes(this->time, eventCount++, record, patient_id, p2_id, purp_id));
    }

    switch (uniform_int(0,3)){
      case 0: 
	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'Treatment'" <<", '" 
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      case 1: 
	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'Payment'" <<", '" 
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
     case 2: 
	if (uniform_double() < purpIn){
	  stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  purp_id << "\")', "
	       << "'Healthcare-op'" <<", '" 
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      }
	
      if (uniform_double() < obtainedConsent ){
	int consentTime = uniform_int(0, this-> time);
	stmt.reset();
	stmt << "INSERT INTO obtain_consent VALUES (" 
	     << "'(str2prin \"_hospital" << Hospital << "\")' , "
	     << "'(str2prin \"_p" <<  patient_id << "\")', "
	     << "'(str2purp \"_pp" <<  purp_id << "\")', '"
	     << intTimeToString (consentTime) << "');";
      run_no_result(stmt);}
      else { 
	int consentTime = uniform_int(0, this-> time);
	stmt.reset();
	stmt << "INSERT INTO obtain_consent VALUES (" 
	     << "'(str2prin \"_fake" << Hospital << "\")' , "
	     << "'(str2prin \"_p" <<  patient_id << "\")', "
	     << "'(str2purp \"_pp" <<  purp_id << "\")', '"
	     << intTimeToString (consentTime) << "');";
      run_no_result(stmt);}
  }

  int nextTime = this->time + exp_int(meanGapConsent);
  addEvent(new DiscloseWithConsent(nextTime, eventCount ++));
}

void Marketing::fire (){
  //508

  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    int rec_no = uniform_int(0, nRecords);
    rec_t record = (*activeRecords[patient_no])[rec_no];
    
    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    // if it is a psychotherapy note then...
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotesWithAuth(this->time, eventCount++, record, patient_id, p2_id, purp_id));
    }

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'marketing'" <<",'" 
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}

    switch(uniform_int(0,3)){
    case 0: 
      if (uniform_double() < obtainedAuthorization ){
	int authTime = uniform_int(0, this->time);
	int msgId = msgCounter ++;
	// insert send p p1 m' u'
	stmt.reset();
	stmt << "INSERT INTO send VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2msg \"_m" << msgId << "\")','"
	     << intTimeToString (authTime)
	     << "');";
	run_no_result(stmt);
	if((uniform_double() <isValidAuth) && genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_is_valid_auth VALUES ("
		 << "'(str2msg \"_m" << msgId << "\")',"
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";}}
      else {
	int authTime = uniform_int(0, this->time);
	int msgId = msgCounter ++;
	// insert send p p1 m' u'
	stmt.reset();
	stmt << "INSERT INTO send VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2prin \"_fake" << Hospital << "\")',"
	     << "'(str2msg \"_m" << msgId << "\")','"
	     << intTimeToString (authTime)
	     << "');";} 
      break;
    case 1: 
      if ((uniform_double()< faceToFace) && genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_face_to_face_comm VALUES ("
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (this->time)
	       << "');";}; 
      break;
      case 2: 
	if ((uniform_double()< promotionalGiftOfNominalVal)&& genSubjPred ){
	    stmt.reset();
	    stmt << "INSERT INTO subj_promotional_gift_of_nominal_val VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";}; 
	break;
    }}
  int nextTime = this->time + exp_int(meanGapMarketing);
  addEvent(new Marketing(nextTime, eventCount ++));
}

void DiscloseToMedicalExaminer::fire (){
  // call body
  charbuf stmt(1024);
  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++; // receipient of the message
  
    //randomly pick a record

  int nRecords = activeRecords[this -> patient_no] -> length();
  
  if (nRecords > 0) {
    int rec_no = uniform_int(0, nRecords);
    rec_t rec = (*activeRecords[this -> patient_no])[rec_no];
    
   stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << rec.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
 
    addEvent(new DiscloseToMedicalExaminerBody(this->time, eventCount++, this->patient_id, p2_id, purp_id));
      }}

void DiscloseToMedicalExaminerBody::fire (){
  
  charbuf stmt(1024);
   switch (uniform_int(0,2)){
    case 0: 
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  this -> receiver_id << "\")', "
	     << "'coroner','"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
       else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  receiver_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 1:
      if (uniform_double() < inRole) {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" << this-> receiver_id << "\")', "
	     << "'medical-examiner','"
	       << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
       else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  receiver_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    };
    
    switch (uniform_int(0,3)){
    case 0: 
      if(uniform_double() < purpIn){
	stmt.reset();
	  stmt << "INSERT INTO purp_in_db VALUES (" 
	       << "'(str2purp \"_pp" <<  this-> purp_id << "\")', "
	       << "'(determine-cause-of-death (str2prin \"_p"
	         << this->patient_id <<"\"))', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
      else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 1:
      if(uniform_double() < purpIn){
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" << this-> purp_id << "\")', "
	     << "'(id (str2prin \"_p" << this->patient_id <<"\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt); }
      else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 2:
      if((uniform_double() < authorizedByLaw) && genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_authorized_by_law VALUES (" 
	       << "'(str2prin \"_rv" << receiver_id <<"\")',"
	       << "'(str2purp \"_pp" << this-> purp_id << "\")', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt); }};
    }

void DiscloseToAssistTPTop:: fire(){
  
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nextDiscloseTime = this->time + exp_int(meanGapDiscloseToAssistTP);
  addEvent(new DiscloseToAssistTP(this->time, eventCount++, patient_no, patient_id));
  addEvent(new DiscloseToAssistTPTop(nextDiscloseTime, eventCount++));
}

void DiscloseToAssistTP::fire(){
  //510/b/1i
  
  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int nRecords = activeRecords[this->patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[this->patient_no])[nRecords-1];

    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
    
  // if it is a psychotherapy note then...
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotesWithAuth(this->time, eventCount++, record, this-> patient_id, p2_id, purp_id));
    }


    switch(uniform_int(0,4)){
    case 0:
      if (uniform_double() < inRole){
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(family-member (str2prin \"_p" << this->patient_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
       else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 1:
      if (uniform_double() < inRole){
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(relative (str2prin \"_p" << this->patient_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
       else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 2:
      if (uniform_double() < inRole){
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(close-personal-friend (str2prin \"_p" << this-> patient_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
       else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 3:
      if (uniform_double() < inRole){
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(identified-164-510b (str2prin \"_p" << this->patient_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
       else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    }

    if((uniform_double() <relevantToInvolvement) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_relevant_to_involvement VALUES ("
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << this->patient_id << "\")', '"
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);}
      
      // b2, b3
      
      if(incapable[patient_no]){
	if ((uniform_double() < inCapableToMakeHCDecisions)&& genSubjPred){
	  stmt.reset();
	stmt << "INSERT INTO subj_Not_Has_capacity_to_make_health_care_decisions VALUES (" 
	     << "'(str2prin \"_p" <<  this -> patient_id << "\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	
	
	if((uniform_double() < professionalJudgmentIsInBestInterestOf510b3)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_professional_judgment_is_in_best_interest_of_164_510_b3 VALUES ("
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << this->patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (this->time)
	       << "');";
	  run_no_result(stmt);} 
	
	if((uniform_double() <relevantToInvolvement)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_relevant_to_involvement VALUES ("
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << this->patient_id << "\")', '"
	       << intTimeToString (this->time)
	       << "');";
	  run_no_result(stmt);};}
      else{
	if ((uniform_double() < CapableToMakeHCDecisions)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_Has_capacity_to_make_health_care_decisions VALUES (" 
	       << "'(str2prin \"_p" <<  this->patient_id << "\")', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	
	switch(uniform_int(0,2)){
	case 0: 
	  if ((uniform_double() < available)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_available VALUES (" 
		 << "'(str2prin \"_p" <<  this->patient_id << "\")', "
		 << "'(str2prin \"_hospital" <<  Hospital << "\")', '"
		 << intTimeToString (this -> time) << "');";
	    run_no_result(stmt);}
	  
	  if((uniform_double() < obtainedAgreement510b2)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_obtained_agreement_164_510_b2 VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << this->patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);} 
	  
	  if((uniform_double() < hasProvidedOppToObj510b2)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_has_provided_opportunity_to_objection_164_510_b2 VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << this->patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);} 
	  
	  if((uniform_double() < professionalJudgmentIndividualDoesNotObject)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_professional_judgment_individual_does_not_object VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << this->patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);} 
	  
	  break;
	case 1: 
	  if ((uniform_double() < notAvailable)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_not_available VALUES (" 
		 << "'(str2prin \"_p" <<  patient_id << "\")', "
		 << "'(str2prin \"_hospital" <<  Hospital << "\")', '"
		 << intTimeToString (this -> time) << "');";
	    run_no_result(stmt);}
	  
	  if((uniform_double() < professionalJudgmentIsInBestInterestOf510b3) && genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_professional_judgment_is_in_best_interest_of_164_510_b3 VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << this->patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);} 
	  
	  if((uniform_double() <relevantToInvolvement)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_relevant_to_involvement VALUES ("
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << this->patient_id << "\")', '"
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);}; 
	  break;}
      }}}

void DiscloseForNotificationTop:: fire(){
  
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nextDiscloseTime = this->time + exp_int(meanGapDiscloseForNotification);
  addEvent(new  DiscloseForNotification(this->time, eventCount++, patient_no, patient_id));
  addEvent(new  DiscloseForNotificationTop(nextDiscloseTime, eventCount++));
}

void DiscloseForNotification::fire(){
 // 510/b/1/ii
  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int nRecords = activeRecords[this ->patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[patient_no])[nRecords-1];

     // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

 // if it is a psychotherapy note then...
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotesWithAuth(this->time, eventCount++, record, this-> patient_id, p2_id, purp_id));
    }

    switch(uniform_int(0,2)){
    case 0:  
      if (uniform_double() < purpIn){
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'assist-notification-164-510b', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      break;
    case 1:
      if (uniform_double() < purpIn){
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'notification-164-510b', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      
      if(uniform_double() < attrIn){
	stmt.reset();
	stmt << "INSERT INTO attr_in_db VALUES ("
	  << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'location-condition-death', '"
	 << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      else {
	stmt.reset();
	stmt << "INSERT INTO attr_in_db VALUES ("
	  << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2attr \"fake\")', '"
	 << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}}
      
      switch(uniform_int(0,3)){
      case 0:
	if (uniform_double() < inRole){
	  stmt.reset();
	  stmt << "INSERT INTO inrole_db VALUES (" 
	       << "'(str2prin \"_rv" <<  p2_id << "\")', "
	       << "'(family-member (str2prin \"_p" << this -> patient_id << "\"))', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	 else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      case 1:
	if (uniform_double() < inRole){
	  stmt.reset();
	  stmt << "INSERT INTO inrole_db VALUES (" 
	       << "'(str2prin \"_rv" <<  p2_id << "\")', "
	       << "'(personal-rep (str2prin \"_p" << this ->patient_id << "\"))', '"
	       << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	 else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      case 2:
	if (uniform_double() < inRole){
	  stmt.reset();
	  stmt << "INSERT INTO inrole_db VALUES (" 
	       << "'(str2prin \"_rv" <<  p2_id << "\")', "
	       << "'(responsible-for-care (str2prin \"_p" << this ->patient_id << "\"))', '"
	     << intTimeToString (this -> time) << "');";
	  run_no_result(stmt);}
	 else {
	stmt.reset();
	stmt << "INSERT INTO inrole_db VALUES (" 
	     << "'(str2prin \"_rv" <<  p2_id << "\")', "
	     << "'(str2role \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	break;
      }
      
 // b2, b3
	if(incapable[patient_no]){
	  if ((uniform_double() < inCapableToMakeHCDecisions) && genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_Not_Has_capacity_to_make_health_care_decisions VALUES (" 
		 << "'(str2prin \"_p" <<  patient_id << "\")', '"
	     << intTimeToString (this -> time) << "');";
	    run_no_result(stmt);}

	  if((uniform_double() < professionalJudgmentIsInBestInterestOf510b3) && genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_professional_judgment_is_in_best_interest_of_164_510_b3 VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << this ->patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);} }
	
	else{
	  if ((uniform_double() < CapableToMakeHCDecisions) && genSubjPred){
	    stmt.reset();
      stmt << "INSERT INTO subj_Has_capacity_to_make_health_care_decisions VALUES (" 
	   << "'(str2prin \"_p" <<  this ->patient_id << "\")', '"
	   << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
	  
	  switch(uniform_int(0,2)){
	  case 0: 
	    if ((uniform_double() < available) && genSubjPred){
	      stmt.reset();
	      stmt << "INSERT INTO subj_available VALUES (" 
		   << "'(str2prin \"_p" <<  this ->patient_id << "\")', "
		   << "'(str2prin \"_hospital" <<  Hospital << "\")', '"
		   << intTimeToString (this -> time) << "');";
	      run_no_result(stmt);}
	    
	    if((uniform_double() < obtainedAgreement510b2) && genSubjPred){
	      stmt.reset();
	      stmt << "INSERT INTO subj_obtained_agreement_164_510_b2 VALUES ("
		   << "'(str2prin \"_hospital" << Hospital << "\")',"
		   << "'(str2prin \"_rv" << p2_id << "\")',"
		   << "'(str2prin \"_p" << this ->patient_id << "\")',"
		   << "'(str2attr \"_t" << record.attr_no << "\")',"
		   << "'(str2purp \"_pp" << purp_id << "\")', '"	
		   << intTimeToString (this->time)
		   << "');";
	      run_no_result(stmt);} 
	    
	    if((uniform_double() < hasProvidedOppToObj510b2) && genSubjPred){
	      stmt.reset();
	      stmt << "INSERT INTO subj_has_provided_opportunity_to_objection_164_510_b2 VALUES ("
		   << "'(str2prin \"_hospital" << Hospital << "\")',"
		   << "'(str2prin \"_rv" << p2_id << "\")',"
		   << "'(str2prin \"_p" << this ->patient_id << "\")',"
		   << "'(str2attr \"_t" << record.attr_no << "\")',"
		   << "'(str2purp \"_pp" << purp_id << "\")', '"	
		   << intTimeToString (this->time)
		   << "');";
	      run_no_result(stmt);} 
	    
	    if((uniform_double() < professionalJudgmentIndividualDoesNotObject) && genSubjPred){
	      stmt.reset();
	      stmt << "INSERT INTO subj_professional_judgment_individual_does_not_object VALUES ("
		   << "'(str2prin \"_hospital" << Hospital << "\")',"
		   << "'(str2prin \"_rv" << p2_id << "\")',"
		   << "'(str2prin \"_p" << this ->patient_id << "\")',"
		   << "'(str2attr \"_t" << record.attr_no << "\")',"
		   << "'(str2purp \"_pp" << purp_id << "\")', '"	
		   << intTimeToString (this->time)
		   << "');";
	      run_no_result(stmt);} 
	    
	    break;
	  case 1: 
	    if ((uniform_double() < notAvailable) && genSubjPred){
	      stmt.reset();
	      stmt << "INSERT INTO subj_not_available VALUES (" 
		   << "'(str2prin \"_p" <<  this ->patient_id << "\")', "
		   << "'(str2prin \"_hospital" <<  Hospital << "\")', '"
		   << intTimeToString (this -> time) << "');";
	      run_no_result(stmt);}
	    
	    if((uniform_double() < professionalJudgmentIsInBestInterestOf510b3) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_professional_judgment_is_in_best_interest_of_164_510_b3 VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << this ->patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
	    break;}}
  }}

void DiscloseToIdCriminal::fire(){
  //512/j/1/iiA
  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[patient_no])[nRecords-1];

     // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
    
   
 // if it is a psychotherapy note then...
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotesWithAuth(this->time, eventCount++, record, patient_id, p2_id, purp_id));
    }

   if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'(id-apprehend (str2prin \"_p" <<  patient_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
   else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
   
   if ((uniform_double() < belongRole) && genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_belongrole VALUES (" 
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'law-enforcement-official', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}

   if ((uniform_double() <consistentWithLaw) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_consistent_with_law VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);}; 

     if(uniform_double() < admittedCrime){
       int adTime = uniform_int(0, this->time);
       int msgId = msgCounter ++;
       // insert send p p1 m' u'
       stmt.reset();
       stmt << "INSERT INTO send VALUES ("
	   << "'(str2prin \"_p" << patient_id << "\")',"
	    << "'(str2prin \"_hospital" << Hospital << "\")',"
	    << "'(str2msg \"_m" << msgId << "\")','"
	    << intTimeToString (adTime)
	    << "');";
       run_no_result(stmt);

       if (genSubjPred){
	 stmt.reset();
	 stmt << "INSERT INTO subj_is_admission_of_crime VALUES ("
	      << "'(str2msg \"_m" << msgId << "\")',"
	      << "'(str2prin \"_p" << patient_id << "\")', '"
	      << intTimeToString (this-> time)
	      << "');";
	 run_no_result(stmt);
       }}
     else {
       int adTime = uniform_int(0, this->time);
       int msgId = msgCounter ++;
       // insert send p p1 m' u'
       stmt.reset();
       stmt << "INSERT INTO send VALUES ("
	   << "'(str2prin \"_p" << patient_id << "\")',"
	    << "'(str2prin \"_fake" << Hospital << "\")',"
	    << "'(str2msg \"_m" << msgId << "\")','"
	    << intTimeToString (adTime)
	    << "');";
       run_no_result(stmt);};

     if ((uniform_double() < believesCrimeCausedHarm) && genSubjPred){
       stmt.reset();
       stmt << "INSERT INTO subj_believes_crime_caused_harm VALUES (" 
	    << "'(str2prin \"_hospital" <<  Hospital << "\")', "
	    << "'(msg (str2info \"_i" << record.info_no <<  ") " 
	    << "(str2purp \"_pp" << purp_id << "\"))', '"
	 //	    << "'(str2msg \"_m" <<  msg_id << "\")', '"
	    << intTimeToString (this -> time) << "');";
       run_no_result(stmt);};

     if ((uniform_double() < notLearnedWhileTreatingPropensityForCrime) && genSubjPred){
       stmt.reset();
       stmt << "INSERT INTO subj_not_learned_while_treating_propensity_for_crime VALUES (" 
	    << "'(str2prin \"_hospital" <<  Hospital << "\")', "
	    << "'(str2prin \"_p" <<  patient_id << "\")', "
	    << "'(str2attr \"_t" <<  record.attr_no << "\")', '"
	    << intTimeToString (this -> time) << "');";
       run_no_result(stmt);};

     if ((uniform_double() < notLearnedThroughRequestForTreatingPropensityForCrime) && genSubjPred){
       stmt.reset();
       stmt << "INSERT INTO subj_not_learned_through_request_for_treating_propensity_for_crime VALUES (" 
	    << "'(str2prin \"_hospital" <<  Hospital << "\")', "
	    << "'(str2prin \"_p" <<  patient_id << "\")', "
	    << "'(str2attr \"_t" <<  record.attr_no << "\")', '"
	    << intTimeToString (this -> time) << "');";
       run_no_result(stmt);};

   if(uniform_double() < attrIn512f2i){
     stmt.reset();
     stmt << "INSERT INTO attr_in_db VALUES ("
	  << "'(str2attr \"_t" << record.attr_no << "\")',"
	  << "'info-164-512-f2i', '"
	 << intTimeToString (this -> time) << "');";
     run_no_result(stmt);}
   else {
     stmt.reset();
     stmt << "INSERT INTO attr_in_db VALUES ("
	  << "'(str2attr \"_t" << record.attr_no << "\")',"
	  << "'(atr2attr \"fake\")', '"
	 << intTimeToString (this -> time) << "');";
     run_no_result(stmt);};
  }
 int nextTime = this->time + exp_int(meanGapIdCriminal);
  addEvent(new DiscloseToIdCriminal(nextTime, eventCount ++));

}

void DiscloseToApprehendEscaped::fire(){
  //512/j/1/iiB

  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[patient_no])[nRecords-1];

    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if ((uniform_double() < belongRole) && genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_belongrole VALUES (" 
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'law-enforcement-official', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}

 // if it is a psychotherapy note then...
    if (record.psychoNotes){
      addEvent(new DisclosePsychotherapyNotesWithAuth(this->time, eventCount++, record, patient_id, p2_id, purp_id));}


    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'(id-apprehend (str2prin \"_p" <<  patient_id << "\"))', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
      

    if ((uniform_double() <consistentWithLaw) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_consistent_with_law VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";}; 

    if ((uniform_double() < believesEscapedLawfulCustody) && genSubjPred ){
       stmt.reset();
       stmt << "INSERT INTO subj_believes_escaped_lawful_custody VALUES (" 
	    << "'(str2prin \"_hospital" << Hospital << "\")', "
	    << "'(str2prin \"_p" <<  patient_id << "\")', '"
	    << intTimeToString (this -> time) << "');";
       run_no_result(stmt);}
  }
 int nextTime = this->time + exp_int(meanGapApprehendEscaped);
  addEvent(new DiscloseToApprehendEscaped(nextTime, eventCount ++));
}

void DiscloseToLawEnforcement::fire (){
  // 512 f1
  int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[patient_no])[nRecords-1];
 
    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
    
    if ((uniform_double() < belongRole) && genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_belongrole VALUES (" 
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'law-enforcement-official', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'law-enforcement', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    switch(uniform_int(0,4)){
    case 0:
      if ((uniform_double() < requiredByLaw)&& genSubjPred ){
	stmt.reset();
	stmt << "INSERT INTO subj_required_by_law VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      break;
    case 1:
      if((uniform_double() < inCompWithCourtOrder) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_in_comp_with_court_order VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      break;
    case 2:
      if((uniform_double() < inCompWithJurySubpoena) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_in_comp_with_subpoena VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      break;
    case 3:
      if((uniform_double() < inCompWithAdminReq) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_in_comp_with_admin_req VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);}
 
      if((uniform_double() < minimumNecessary) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_minimum_necessary VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(msg (str2info \"_i" << record.info_no <<  "\")" 
	     << "(str2purp \"_pp" << purp_id << "\"))', "
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      
      if ((uniform_double() < deidentifiedInfoNotSufficient) && genSubjPred ){
	stmt.reset();
	stmt << "INSERT INTO subj_deidentified_info_not_sufficient VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")',' "
	    << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    }}
  
  int nextTime = this->time + exp_int(meanGapToLaw);
  addEvent(new DiscloseToLawEnforcement(nextTime, eventCount ++));
}


void LawEnforcementRequestForIdLoc::fire(){
  //512/f2
  // could use a different send

  int req_msg_id = msgCounter ++;
  int p2_id = principalCounter ++;
  charbuf stmt(1024);

  // insert send
  stmt.reset();
  stmt << "INSERT INTO send VALUES ("
       << "'(str2prin \"_rv" << p2_id << "\")',"
       << "'(str2prin \"_hospital" << Hospital << "\")',"
       << "'(str2msg \"_m" << req_msg_id << "\")','"
       << intTimeToString (this->time)
       << "');";
  run_no_result(stmt);

  int responseTime = this -> time + exp_int(meanGapResponse);
  int purp_id = purposeCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {

    rec_t record = (*activeRecords[patient_no])[nRecords-1];

    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if ((uniform_double() < belongRole) && genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_belongrole VALUES (" 
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'law-enforcement-official', '"
	   << intTimeToString (responseTime) << "');";
      run_no_result(stmt);}
    
    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'(law-enforcement-relevant-id-or-location (str2prin \"_p"
	   <<  patient_id << "\"))', '"
	   << intTimeToString (responseTime) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    if((uniform_double() <isRequest) && genSubjPred) {
	stmt.reset();
	stmt << "INSERT INTO subj_is_request VALUES ("
	     << "'(str2msg \"_m" << req_msg_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";}

     if(uniform_double() < attrIn512f2i){
     stmt.reset();
     stmt << "INSERT INTO attr_in_db VALUES ("
	  << "'(str2attr \"_t" << record.attr_no << "\")',"
	  << "'info-164-512-f2i', '"
	 << intTimeToString (this -> time) << "');";
     run_no_result(stmt);}
     else {
     stmt.reset();
     stmt << "INSERT INTO attr_in_db VALUES ("
	  << "'(str2attr \"_t" << record.attr_no << "\")',"
	  << "'(str2attr \"fake\")', '"
	 << intTimeToString (this -> time) << "');";
     run_no_result(stmt);}
  }
  
  int nextTime = this->time + exp_int(meanGapLawReq);
  addEvent(new LawEnforcementRequestForIdLoc(nextTime, eventCount ++));
}


void LawEnforcementRequestVictim::fire(){
  //512/f3
  // could use a different send

  int req_msg_id = msgCounter ++;
  int p2_id = principalCounter ++;
  charbuf stmt(1024);

  // insert send
  stmt.reset();
  stmt << "INSERT INTO send VALUES ("
       << "'(str2prin \"_rv" << p2_id << "\")',"
       << "'(str2prin \"_hospital" << Hospital << "\")',"
       << "'(str2msg \"_m" << req_msg_id << "\")','"
       << intTimeToString (this->time)
       << "');";
  run_no_result(stmt);

  int responseTime = this -> time + exp_int(meanGapResponse);
  int purp_id = purposeCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {

    rec_t record = (*activeRecords[patient_no])[nRecords-1];
 
    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if ((uniform_double() < belongRole) && genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_belongrole VALUES (" 
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'law-enforcement-official', '"
	   << intTimeToString (responseTime) << "');";
      run_no_result(stmt);}
    
    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'law-enforcement', '"
	   << intTimeToString (responseTime) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    if((uniform_double() <isRequest) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_is_request VALUES ("
	     << "'(str2msg \"_m" << req_msg_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";}

    switch(uniform_int(0,2)){
    case 0:
      if ((uniform_double() < belongRole) && genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_belongrole VALUES (" 
	     << "'(str2prin \"_p" <<  patient_id << "\")', "
	     << "'victim-of-crime', '"
	     << intTimeToString (responseTime) << "');";
	run_no_result(stmt);}
      break;
    case 1:
      if ((uniform_double() < belongRole)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_belongrole VALUES (" 
	     << "'(str2prin \"_p" <<  patient_id << "\")', "
	     << "'suspected-victim-of-crime', '"
	     << intTimeToString (responseTime) << "');";
	run_no_result(stmt);}
      break;
    }

    switch(uniform_int(0,2)){
    case 0:
      if(uniform_double() < obtainedAgreement){
	int agreeTime = uniform_int(this->time, responseTime);
	int agree_msgId = msgCounter ++;
	// insert send p p1 m' u'
	stmt.reset();
	stmt << "INSERT INTO send VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2msg \"_m" << agree_msgId << "\")','"
	     << intTimeToString (agreeTime)
	     << "');";
	run_no_result(stmt);
	if((uniform_double() <isAgreement)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_is_valid_agreement VALUES ("
	       << "'(str2msg \"_m" << agree_msgId << "\")',"
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (responseTime)
	       << "');";
	  run_no_result(stmt);}}
      else {
	int agreeTime = uniform_int(this->time, responseTime);
	int agree_msgId = msgCounter ++;
	// insert send p p1 m' u'
	stmt.reset();
	stmt << "INSERT INTO send VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2prin \"_fake" << Hospital << "\")',"
	     << "'(str2msg \"_m" << agree_msgId << "\")','"
	     << intTimeToString (agreeTime)
	     << "');";
	run_no_result(stmt);}
      break;
    case 1:

      if((uniform_double() < isEmergency)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_is_emergency VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")', '"
	     << intTimeToString (responseTime)
	     << "');";
	run_no_result(stmt);}

      if((uniform_double() <representsNeededToDetermineCrime)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_represents_needed_to_determine_crime VALUES ("
	  << "'(str2prin \"_p" << p2_id << "\")',"
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (responseTime)
	       << "');";
	  run_no_result(stmt);}

      if((uniform_double() <representsNotUsedAgainstVictim )&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_represents_not_used_against_victim VALUES ("
	  << "'(str2prin \"_p" << p2_id << "\")',"
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (responseTime)
	       << "');";
	  run_no_result(stmt);}
      if((uniform_double() <representsActivityAdverselyAffectedByWaiting )&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_represents_activity_adverserly_affected_by_waiting VALUES ("
	  << "'(str2prin \"_p" << p2_id << "\")',"
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (responseTime)
	       << "');";
	  run_no_result(stmt);}

      if((uniform_double() <believesInBestInterest )&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_believes_in_best_interest VALUES ("
	  << "'(str2prin \"_p" << p2_id << "\")',"
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (responseTime)
	       << "');";
	  run_no_result(stmt);}
      break;
    }
  }
 int nextTime = this->time + exp_int(meanGapLawReq);
 addEvent(new LawEnforcementRequestVictim(nextTime, eventCount ++));
}

void ReportAbuse::fire (){
  //512c
 int purp_id = purposeCounter ++;
  int p2_id = principalCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {
    charbuf stmt(1024);
    rec_t record = (*activeRecords[patient_no])[nRecords-1];
 
    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);
    
    if ((uniform_double() < belongRole)&& genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_belongrole VALUES (" 
	   << "'(str2prin \"_rv" <<  p2_id << "\")', "
	   << "'gov-authority', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}

    if((uniform_double() < authorizedByLawToRecv)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_authorized_by_law_to_recv VALUES ("
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);}

    if((uniform_double() < believesVictimOfAbuse)&& genSubjPred){
      stmt.reset();
      stmt << "INSERT INTO subj_believes_victim_of_abuse VALUES ("
	   << "'(str2prin \"_hospital" << Hospital << "\")',"
	   << "'(str2prin \"_p" << patient_id << "\")', '"
	   << intTimeToString (this->time)
	   << "');";
      run_no_result(stmt);} 

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'reports-of-abuse', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    switch(uniform_int(0,3)){
    case 0:
      if((uniform_double() < requiredByLaw)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_required_by_law VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      break;
    case 1:
        if(uniform_double() < hasAgreed){
	stmt.reset();
	stmt << "INSERT INTO individual_has_agreed VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);}
	break;
    case 2:
      if((uniform_double() < authorizedByStatuteOrRegulation)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_authorized_by_statute_or_regulation VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);}
      if(incapable[patient_no]){
	if((uniform_double() < assuranceDisclosureNotUsedAgainstIndividual)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_assurance_disclosure_not_used_against_individual VALUES ("
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	       << "');";
	run_no_result(stmt);}
	if((uniform_double() < believesWaitingForAgreementWouldHinderEnforcement)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_believes_waiting_for_agreement_would_hinder_enforcement VALUES ("
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_rv" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	       << "');";
	run_no_result(stmt);}
      }
      else{ 
	if((uniform_double() < believesDisclosureNecessaryToPreventHarm)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_believes_disclosure_necessary_to_prevent_harm VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);}
      }
	break;
    }

    switch(uniform_int(0,4)){
    case 0:
      if(uniform_double() < sendNotice){
	int noticeTime = uniform_int(this->time, this->time + prompt);
	int notice_msgId = msgCounter ++;
	// insert send p p1 m' u'
	stmt.reset();
	stmt << "INSERT INTO send VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2msg \"_m" << notice_msgId << "\")','"
	     << intTimeToString (noticeTime)
	     << "');";
	run_no_result(stmt);
	if((uniform_double() <isNoticeOfReport)&& genSubjPred){
	  stmt.reset();
	  stmt << "INSERT INTO subj_is_notice_of_report VALUES ("
	       << "'(str2msg \"_m" << notice_msgId << "\")',"
	       << "'(str2prin \"_hospital" << Hospital << "\")',"
	       << "'(str2prin \"_p" << p2_id << "\")',"
	       << "'(str2prin \"_p" << patient_id << "\")',"
	       << "'(str2attr \"_t" << record.attr_no << "\")',"
	       << "'(str2purp \"_pp" << purp_id << "\")', '"	
	       << intTimeToString (noticeTime)
	       << "');";
	  run_no_result(stmt);}}
      else{
	int noticeTime = uniform_int(this->time, this->time + prompt);
	int notice_msgId = msgCounter ++;
	// insert send p p1 m' u'
	stmt.reset();
	stmt << "INSERT INTO send VALUES ("
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2prin \"_fake" << Hospital << "\")',"
	     << "'(str2msg \"_m" << notice_msgId << "\")','"
	     << intTimeToString (noticeTime)
	     << "');";}
    case 1: 
      if((uniform_double() < believesNoticeWouleRiskIndividual)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_believes_notice_would_risk_indivdual VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_p" << p2_id << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);}
      break;
    case 2: 
      if((uniform_double() < believesInformingIsNotInTheBestInterest)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_believes_informing_is_not_in_the_best_interest VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_p" << p2_id << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);}
      break;
    case 3: 
      if((uniform_double() < wouldInformPr)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_would_inform_pr VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_rv" << p2_id << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")',"
		 << "'(str2attr \"_t" << record.attr_no << "\")',"
		 << "'(str2purp \"_pp" << purp_id << "\")', '"	
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);}
      if((uniform_double() < believesPrResponsible)&& genSubjPred){
	    stmt.reset();
	    stmt << "INSERT INTO subj_believes_pr_responsible VALUES ("
		 << "'(str2prin \"_hospital" << Hospital << "\")',"
		 << "'(str2prin \"_p" << patient_id << "\")', '"
		 << intTimeToString (this->time)
		 << "');";
	    run_no_result(stmt);}
	   break;
    }
  }
 int nextTime = this->time + exp_int(meanGapReportAbuse);
  addEvent(new ReportAbuse(nextTime, eventCount ++));
}


void CourtOrder::fire (){
  // 512/e/1/i 
  int p2_id = principalCounter ++;
  int purp_id = purposeCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  charbuf stmt(1024);
  
  if (nRecords > 0) {

    rec_t record = (*activeRecords[patient_no])[nRecords-1];
    //insert m = msg(info, purpose)
 
    // insert send
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'judicial-admin-proceeding', '"
	   << intTimeToString (this -> time) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    

    switch(uniform_int(0,2)){
    case 0: 
      if((uniform_double() < authorizedByCourtOrder)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_authorized_by_court_order VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      break;

    case 1: 
      if((uniform_double() < authorizedByTribunalOrder)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_authorized_by_tribunal_order VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (this->time)
	     << "');";
	run_no_result(stmt);} 
      break;
  }}
  int nextTime = this->time + exp_int(meanGapCourtOrder);
  addEvent(new CourtOrder(nextTime, eventCount ++));
}

void LegalProceeding:: fire(){
 // 512/e/1/ii
  // could use a different send

  int req_msg_id = msgCounter ++;
  int p2_id = principalCounter ++;
  charbuf stmt(1024);

  // insert send
  stmt.reset();
  stmt << "INSERT INTO send VALUES ("
       << "'(str2prin \"_rv" << p2_id << "\")',"
       << "'(str2prin \"_hospital" << Hospital << "\")',"
       << "'(str2msg \"_m" << req_msg_id << "\")','"
       << intTimeToString (this->time)
       << "');";
  run_no_result(stmt);

  int responseTime = this -> time + exp_int(meanGapResponse);
  int purp_id = purposeCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {

    rec_t record = (*activeRecords[patient_no])[nRecords-1];

    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'judicial-admin-proceeding','"
	   << intTimeToString (responseTime) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    if((uniform_double() <isLawfulProcess)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_is_lawful_process VALUES ("
	     << "'(str2msg \"_m" << req_msg_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";
	run_no_result(stmt);}

    if((uniform_double() < receivedSatisfactoryAssurance512e1iii)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_received_satisfactory_assurance_164_512e1iii VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";
	run_no_result(stmt);}

    if((uniform_double() < receivedSatisfactoryAssurance512e1iv)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_received_satisfactory_assurance_164_512e1iv VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";
	run_no_result(stmt);}
  }
  int nextTime = this->time + exp_int(meanGapLawReq);
  addEvent(new LegalProceeding(nextTime, eventCount ++));
}


void DiscloseInLawfulProcessMadeEffortToNotify:: fire(){
  // 512/e/1/vi
  // could use a different send

  int req_msg_id = msgCounter ++;
  int p2_id = principalCounter ++;
  charbuf stmt(1024);

  // insert send
  stmt.reset();
  stmt << "INSERT INTO send VALUES ("
       << "'(str2prin \"_rv" << p2_id << "\")',"
       << "'(str2prin \"_hospital" << Hospital << "\")',"
       << "'(str2msg \"_m" << req_msg_id << "\")','"
       << intTimeToString (this->time)
       << "');";
  run_no_result(stmt);

  int responseTime = this -> time + exp_int(meanGapResponse);
  int purp_id = purposeCounter ++;
  int patient_no = uniform_int(0, nPatient);
  int patient_id = currPatients[patient_no];
  int nRecords = activeRecords[patient_no] -> length();
  
  if (nRecords > 0) {

    rec_t record = (*activeRecords[patient_no])[nRecords-1];

    // insert disclose
    stmt.reset();
    stmt << "INSERT INTO disclose VALUES ("
	 << "'(str2prin \"_hospital" << Hospital << "\")',"
	 << "'(str2prin \"_rv" << p2_id << "\")',"
	 << "'(msg (str2info \"_i" << record.info_no <<  "\") " 
	 << "(str2purp \"_pp" << purp_id << "\"))', '"
	 << intTimeToString (this->time)
	 << "');";
    run_no_result(stmt);

    if (uniform_double() < purpIn){
      stmt.reset();
      stmt << "INSERT INTO purp_in_db VALUES (" 
	   << "'(str2purp \"_pp" <<  purp_id << "\")', "
	   << "'judicial-admin-proceeding','"
	   << intTimeToString (responseTime) << "');";
      run_no_result(stmt);}
    else{
	stmt.reset();
	stmt << "INSERT INTO purp_in_db VALUES (" 
	     << "'(str2purp \"_pp" <<  purp_id << "\")', "
	     << "'(str2purp \"fake\")', '"
	     << intTimeToString (this -> time) << "');";
	run_no_result(stmt);}
    
    if((uniform_double() <isLawfulProcess)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_is_lawful_process VALUES ("
	     << "'(str2msg \"_m" << req_msg_id << "\")',"
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";
	run_no_result(stmt);}

    if((uniform_double() < madeReasonableEffortToNotify)&& genSubjPred){
	stmt.reset();
	stmt << "INSERT INTO subj_made_reasonable_effort_to_notify VALUES ("
	     << "'(str2prin \"_hospital" << Hospital << "\")',"
	     << "'(str2prin \"_rv" << p2_id << "\")',"
	     << "'(str2prin \"_p" << patient_id << "\")',"
	     << "'(str2attr \"_t" << record.attr_no << "\")',"
	     << "'(str2purp \"_pp" << purp_id << "\")', '"	
	     << intTimeToString (responseTime)
	     << "');";
	run_no_result(stmt);}
  }
 int nextTime = this->time + exp_int(meanGapLawReq);
 addEvent(new DiscloseInLawfulProcessMadeEffortToNotify(nextTime, eventCount ++));
}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////



void create_tables()
{

  // We create one table for each trace predicate in our system. The
  // exception is the predicate inrole, which can be defined using
  // beginrole and endrole. The character buffer cmd_buf holds our SQL
  // statements. All first-order terms are represented by their string
  // representations and, therefore, become VARCHAR2 in the
  // database. The only exceptions are time points which are
  // represented as integers wherever possible, for efficiency
  // reasons.

  // We also store in tables starting with the prefix subj_ subjective
  // predicates. This may come in handy if we start simulating a human
  // oracle during audit.

  charbuf cmd_buf(1024); // A 1K buffer should suffice.

  cmd_buf.reset();
  cmd_buf << "BEGIN TRANSACTION;";
  run_no_result(cmd_buf);

  //  cmd_buf.reset();
  //  cmd_buf << "CREATE TABLE timeline_db (time VARCHAR2(50) UNIQUE);";  
  //  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE attr_in_db (subattr VARCHAR2(50), supattr VARCHAR2(50) NOT NULL,time VARCHAR2(50) NOT NULL);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE hasattrof_db (info VARCHAR2(50),  principal VARCHAR2(50) NOT NULL, attribute VARCHAR2(50) NOT NULL,time VARCHAR2(50) NOT NULL);";  
  run_no_result(cmd_buf); 


  
  cmd_buf.reset();
  cmd_buf << "CREATE TABLE send (sender VARCHAR2(50) NOT NULL,  receiver VARCHAR2(50) NOT NULL, message VARCHAR2(50) NOT NULL,time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE disclose (sender VARCHAR2(50) NOT NULL,  receiver VARCHAR2(50) NOT NULL, message VARCHAR2(50) NOT NULL,time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);
  
  cmd_buf.reset();
  cmd_buf << "CREATE TABLE request (sender VARCHAR2(50) NOT NULL,  receiver VARCHAR2(50) NOT NULL, message VARCHAR2(50) NOT NULL,time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_workforce_member"
	  << "(employee VARCHAR2(50) NOT NULL,"
	  << "employer VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE individual_has_agreed"
	  << "(sender VARCHAR2(50) NOT NULL,"
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL," 
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


  cmd_buf.reset();
  cmd_buf << "CREATE TABLE inrelation (patient VARCHAR2(50) NOT NULL, hospital VARCHAR2(50) NOT NULL, relation VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);
  
  cmd_buf.reset();
  cmd_buf << "CREATE TABLE deceased (principal VARCHAR2(50) NOT NULL,"
	  <<  "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE obtain_consent"
	  << "(hospital VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

 cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_participant_of_organized_health (principal VARCHAR2(50) NOT NULL, organizedhealth VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

 cmd_buf.reset();
  cmd_buf << "CREATE TABLE beginrole (principal VARCHAR2(50) NOT NULL,  role VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE endrole (principal VARCHAR2(50) NOT NULL,  role VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

 cmd_buf.reset();
  cmd_buf << "CREATE TABLE inrole_db (principal VARCHAR2(50) NOT NULL,  role VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

 cmd_buf.reset();
  cmd_buf << "CREATE TABLE in_lawful_custody"
	  << "(criminal VARCHAR2(50) NOT NULL,"
	  << "law VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

 // %%%%%%%%%%%%% CREATE TABLES FOR UNUSED OBJECTIVE PREDICATES %%%%%%%%%%%%%%%%%%%%%

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE health_insurance_placed_with (insurance VARCHAR2(50) NOT NULL,  principal VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE directed_disclosure"
	  << "(director VARCHAR2(50) NOT NULL," 
	  << "sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "principal VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL," 
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE FDA_regulated (product VARCHAR2(50) NOT NULL,  time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_business_assoc"
	  << "(business VARCHAR2(50) NOT NULL,"
	  << "principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE provides_medical_surveillance"
	  << "(provider VARCHAR2(50) NOT NULL,"
	  << "principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


  cmd_buf.reset();
  cmd_buf << "CREATE TABLE provides_medical_injury_evaluation"
	  << "(provider VARCHAR2(50) NOT NULL,"
	  << "principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE deemed_necessary_for_mission"
	  << "(authority VARCHAR2(50) NOT NULL," 
	  << "sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "principal VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL," 
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_gov_agency_administering"
	  << "(admin VARCHAR2(50) NOT NULL,"
	  << "principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE determines_eligibility_enrollemnt"
	  << "(authority VARCHAR2(50) NOT NULL,"
	  << "principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE enrollemnt_info_collected_by"
	  << "(principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_health_insurance_issuer"
	  << "(issuer VARCHAR2(50) NOT NULL,"
	  << "principal VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_hmo"
	  << "(principal VARCHAR2(50) NOT NULL,"
	  << "hmo VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE is_victim_of"
	  << "(principal VARCHAR2(50) NOT NULL,"
	  << "crime VARCHAR2(50) NOT NULL,"
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE purp_in_db (subpurpose VARCHAR2(50) NOT NULL,  suppurpose VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


  // %%%%%%%%%%%%% CREATE TABLES FOR SUBJECTIVE PREDICATES %%%%%%%%%%%%%%%%%%%%%
  if (genSubjPred) {

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_pertain_to (attr VARCHAR2(50) NOT NULL, relation VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_begin_incapable (principal VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_end_incapable (principal VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_minimum_necessary"
	  << "(sender VARCHAR2(50) NOT NULL,"
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "message VARCHAR2(50) NOT NULL," 
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_belongrole (principal VARCHAR2(50) NOT NULL,  role VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


  cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_represents_needed_to_determine_crime (law VARCHAR2(50) NOT NULL,  sender VARCHAR2(50) NOT NULL, receiver VARCHAR2(50) NOT NULL , patient VARCHAR2(50) NOT NULL, attribute VARCHAR2(50) NOT NULL,  purpose VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_represents_not_used_against_victim (law VARCHAR2(50) NOT NULL,  sender VARCHAR2(50) NOT NULL, receiver VARCHAR2(50) NOT NULL, patient VARCHAR2(50) NOT NULL, attribute VARCHAR2(50) NOT NULL,  purpose VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_represents_activity_adverserly_affected_by_waiting (law VARCHAR2(50) NOT NULL,  sender VARCHAR2(50) NOT NULL, receiver VARCHAR2(50) NOT NULL,  patient VARCHAR2(50) NOT NULL, attribute VARCHAR2(50) NOT NULL,  purpose VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_believes_in_best_interest (law VARCHAR2(50) NOT NULL,  sender VARCHAR2(50) NOT NULL, receiver VARCHAR2(50) NOT NULL, patient VARCHAR2(50) NOT NULL, attribute VARCHAR2(50) NOT NULL,  purpose VARCHAR2(50) NOT NULL, time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_is_request"
	  << "(message VARCHAR2(50) NOT NULL," 
	  << "sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_is_valid_agreement"
	  << "(message VARCHAR2(50) NOT NULL," 
	  << "sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_authorized_by_court_order"
	  << "(sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_authorized_by_tribunal_order"
	  << "(sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_deidentified_info_not_sufficient"
	  << "(purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_in_comp_with_court_order"
	  << "(sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_in_comp_with_subpoena"
	  << "(sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
  cmd_buf << "CREATE TABLE subj_in_comp_with_admin_req"
	  << "(sender VARCHAR2(50) NOT NULL," 
	  << "receiver VARCHAR2(50) NOT NULL," 
	  << "patient VARCHAR2(50) NOT NULL," 
	  << "attribute VARCHAR2(50) NOT NULL,"
	  << "purpose VARCHAR2(50) NOT NULL," 
	  << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_death_may_be_result_of_crime" 
	 << "(hospital VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_evidence_of_crime_on_premises" 
	 << "(hospital VARCHAR2(50) NOT NULL,"
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_is_valid_auth" 
	 <<" (message VARCHAR2(50) NOT NULL,"
	 << "sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_relevant_to_involvement" 
   	 << "(attribute VARCHAR2(50) NOT NULL,"
	 << "sender VARCHAR2(50) NOT NULL,"
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_Not_Has_capacity_to_make_health_care_decisions"
   	 << "(principal VARCHAR2(50) NOT NULL,"
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_Has_capacity_to_make_health_care_decisions"
   	 << "(principal VARCHAR2(50) NOT NULL,"
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_is_emergency"
   	 << "(principal VARCHAR2(50) NOT NULL,"
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_available"
   	 << "(patient VARCHAR2(50) NOT NULL,"
	 << "hospital VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_not_available"
   	 << "(patient VARCHAR2(50) NOT NULL,"
	 << "hospital VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_obtained_agreement_164_510_b2"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL," 
   	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_has_provided_opportunity_to_objection_164_510_b2"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL," 
   	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_professional_judgment_individual_does_not_object"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL," 
   	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_professional_judgment_is_in_best_interest_of_164_510_b3" 
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL," 
   	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_not_learned_while_treating_propensity_for_crime"
	 << "(hospital VARCHAR2(50) NOT NULL,"
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_not_learned_through_request_for_treating_propensity_for_crime"
	 << "(hospital VARCHAR2(50) NOT NULL,"
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_face_to_face_comm" 
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_promotional_gift_of_nominal_val"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_necessary_for_duties" 
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_is_lawful_process" 
	 <<" (message VARCHAR2(50) NOT NULL,"
	 << "sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_is_notice_of_report"
	 <<" (message VARCHAR2(50) NOT NULL,"
	 << "sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_received_satisfactory_assurance_164_512e1iii"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_received_satisfactory_assurance_164_512e1iv"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_notice_would_risk_indivdual"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_made_reasonable_effort_to_notify"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_informing_is_not_in_the_best_interest"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_would_inform_pr"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_pr_responsible"
	 << "(hospital VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_authorized_by_law" 
	 << "(principal VARCHAR2(50) NOT NULL," 
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_authorized_by_law_to_recv" 
	 << "(principal VARCHAR2(50) NOT NULL," 
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_victim_of_abuse"
	 << "(hospital VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);


cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_consistent_with_law" 
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_required_by_law"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_authorized_by_statute_or_regulation"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_assurance_disclosure_not_used_against_individual"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_waiting_for_agreement_would_hinder_enforcement"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE  subj_believes_disclosure_necessary_to_prevent_harm"
	 << "(sender VARCHAR2(50) NOT NULL,"
	 << "receiver VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "attribute VARCHAR2(50) NOT NULL,"
	 << "purpose VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_crime_caused_harm"
	 << "(principal VARCHAR2(50) NOT NULL," 
	 << "message VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_is_admission_of_crime"
	 << "(message VARCHAR2(50) NOT NULL," 
	 << "principal VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);

cmd_buf.reset();
 cmd_buf << "CREATE TABLE subj_believes_escaped_lawful_custody"
	 << "(hospital VARCHAR2(50) NOT NULL," 
	 << "patient VARCHAR2(50) NOT NULL," 
	 << "time VARCHAR2(50) NOT NULL);";
  run_no_result(cmd_buf);
  }}

void create_index(){

  charbuf cmd_buf(1024); // A 1K buffer should suffice.

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_ATTR on attr_in_db (subattr);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_MSG on disclose (message);";
  run_no_result(cmd_buf);

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_HADATTR on hasattrof_db (info);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_INREL on inrelation (patient, hospital);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_INROLE on inrole_db (principal, role);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_ORG on is_participant_of_organized_health (principal);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_WORK on is_workforce_member (employee);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_PURP on purp_in_db (subpurpose);";  
  run_no_result(cmd_buf); 

  cmd_buf.reset();
  cmd_buf << "CREATE INDEX IDX_SEND on send (sender, receiver);";  
  run_no_result(cmd_buf); 

}
// Initialize the simulation
void init_sim()
{

  create_tables();

  create_index();

  charbuf stmt(1024);

  stmt.reset();

  stmt << "INSERT INTO beginrole VALUES (" 
       << "'(str2prin \"_hospital" <<  Hospital << "\")', "
       << "'covered-entity', '"
       << intTimeToString (0) << "');";

  run_no_result(stmt);

  int i;
  for (i = 0; i < nPatient; i ++) {
    currPatients[i] = -1;
    incapable[i] = -1;
    inrelation[i] = false;
    activeRecords[i] = new boundedbuffer<rec_t> (nActiveRecords);
  }

 for (i = 0; i < nPatient; i++) {
    addEvent(new ReplacePatient(0, eventCount ++, i));
  }

 // addEvent(new LegalOrderForLawEnforcement(0, eventCount ++));
 addEvent(new LawEnforcementRequestForIdLoc(0, eventCount ++));
 addEvent(new LawEnforcementRequestVictim(0, eventCount ++));

 addEvent(new CourtOrder(0, eventCount ++));
 addEvent(new LegalProceeding(0, eventCount ++));
 addEvent(new DiscloseInLawfulProcessMadeEffortToNotify(0, eventCount ++));
 
// 502/e/1
// addEvent(new DiscloseToBusinessAssoc(0, eventCount ++));

 // 506/b/1
 
 addEvent(new DiscloseWithConsent(0, eventCount ++));

 // 508
 addEvent(new Marketing(0, eventCount ++));

 // 510/b/4
 // addEvent(new DiscloseForDisasterRelief(0, eventCount ++));

 // 510/b/1/i
 addEvent(new DiscloseForNotificationTop(0, eventCount ++));

 // 510/b/1/ii
 addEvent(new DiscloseToAssistTPTop(0, eventCount ++));

 // 512/b/1/i
 // addEvent(new DiscloseForPublicHealth(0, eventCount ++));

 // 512/b/1/ii
 // addEvent(new ReportChildAbuse(0, eventCount ++));
 // 512/b/1/iii
 // addEvent(new FDASafety(0, eventCount ++));

 // 512/b/1iv
 // addEvent(new DiscloseForPublicHealthIntervention(0, eventCount ++));
 // 512/c
 addEvent(new ReportAbuse(0, eventCount ++));
 // 512/c
 // addEvent(new DiscloseToHealthOversightAgency(0, eventCount ++));

 // 512/f/1
 addEvent(new DiscloseToLawEnforcement(0, eventCount ++));
  // 512/f/5
 addEvent(new ReportCrimeOnPremise(0, eventCount ++));
 // 512/h
 // addEvent(new DiscloseToOrgan(0, eventCount ++));

 // 512/i
 // addEvent(new DiscloseForResearch(0, eventCount ++));
 // 512/j/1/i
 // addEvent(new DiscloseToLessenRisk(0, eventCount ++));

 // 512/j/1/iiA
 addEvent(new DiscloseToIdCriminal(0, eventCount ++));
 // 512/j/1/iiB
 addEvent(new DiscloseToApprehendEscaped(0, eventCount ++));

 // 512/l
 // addEvent(new DiscloseForWorkerCompensation(0, eventCount ++));
 // 514/e
 // addEvent(new DiscloseWithLimitedDataAgreement(0, eventCount ++));
 // 514/f
 // addEvent(new DiscloseForFundRaising(0, eventCount ++));
 // 514/g
 // addEvent(new DiscloseInsuranceRelatedInfo(0, eventCount ++));
}


// Finalize the simulation
void finalize_sim()
{
  charbuf stmt(1024);
  stmt << "END TRANSACTION;";
  run_no_result(stmt);

  for (int i = 0; i < nPatient; i++) {
    delete (activeRecords[i]);
  }
}




//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////



// Initialize the database 
int init_db()
{
#ifdef USE_DB
  if (sqlite3_open_v2(dbname, &pDb, 
		      SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE,
		      NULL) != SQLITE_OK) {
    cerr << "Failed to open database: " << dbname << endl;
    return 0;
  }
#endif
  return 1;
}

// Close the database
int finalize_db()
{
#ifdef USE_DB
  if (sqlite3_close(pDb) != SQLITE_OK) {
    cerr << "Failed to close database: " << dbname << endl;
    return 0;
  }  
  pDb = NULL;
#endif
  return 1;
}


//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////


int main ()
{
  if (!init_db()) {
    finalize_db();
    return 1;
  }

  // Initialize the random number generator
  srand (1);
  


  run_sim(maxSimTime);

  if (!finalize_db()) {
    return 1;
  }

  return 0;
}
