/*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <unistd.h>
#include <sqlite3.h>
#include <string.h>
#include "sqlite-wrap.h"

Pointer smli_sqlite3_open (Pointer _str)
{
  const char * dbname = (const char *) _str;
  sqlite3 * pDb = NULL;

  int r = sqlite3_open_v2 (dbname, &pDb, SQLITE_OPEN_READONLY, NULL);

  if (r == SQLITE_OK) {
    // The connection opened successfully
    return ((Pointer) pDb);
  }
  else {
    // Connection did not open correctly
    sqlite3_close (pDb);
    return NULL;
  }

}

void smli_sqlite3_close (Pointer _pDb)
{
  sqlite3_close ((sqlite3 *) _pDb);
}

Pointer smli_sqlite3_prepare (Pointer _pDb, Pointer _sql)
{
  sqlite3 * pDb = (sqlite3 *) _pDb;
  const char * sql = (const char *) _sql;

  if (pDb == NULL) return NULL;

  sqlite3_stmt * pStmt = NULL;

  int r = sqlite3_prepare_v2(pDb, sql, -1, &pStmt, NULL);

  if (r == SQLITE_OK) {
    return ((Pointer) pStmt);
  }
  else {
    // Failed to compile
    sqlite3_finalize (pStmt);
    return NULL;
  }
}

Int32 smli_sqlite3_step (Pointer _pStmt)
{
  sqlite3_stmt * pStmt = (sqlite3_stmt *) _pStmt;
  int r = sqlite3_step (pStmt);
  if (r == SQLITE_ROW) {
    return 0;
  }
  else if (r == SQLITE_DONE) {
    return 1;
  }
  else {
    return 2;
  }
}

Int32 smli_sqlite3_column_count (Pointer _pStmt)
{
  sqlite3_stmt * pStmt = (sqlite3_stmt *) _pStmt;
  
  return ((Int32) (sqlite3_column_count(pStmt)));
}

Int32 smli_sqlite3_column_bytes (Pointer _pStmt, Int32 _n)
{
  sqlite3_stmt * pStmt = (sqlite3_stmt *) _pStmt;
  int n = (int) _n;

  return ((Int32) (sqlite3_column_bytes(pStmt, n)));
}

void smli_sqlite3_column_text (Pointer _pStmt, Int32 _n, Pointer _buf)
{
  sqlite3_stmt * pStmt = (sqlite3_stmt *) _pStmt;
  int n = (int) _n;
  char * buf = (char *) _buf;

  const unsigned char * data = sqlite3_column_text(pStmt, n);

  strcpy (buf, (const char *) data);
}

void smli_sqlite3_reset (Pointer _pStmt)
{
  sqlite3_stmt * pStmt = (sqlite3_stmt *) _pStmt;

  sqlite3_reset (pStmt);
}

void smli_sqlite3_finalize (Pointer _pStmt)
{
  sqlite3_stmt * pStmt = (sqlite3_stmt *) _pStmt;

  sqlite3_finalize (pStmt);
}

// implementing the backup api for loading DB into memory
Pointer smli_sqlite3_backup_init (Pointer _pDest, Pointer _zDestName, Pointer _pSource, Pointer _zSourceName)
{
  sqlite3 * pDest = (sqlite3 *) _pDest;
  const char * zDestName = (const char *) _zDestName;
  sqlite3 * pSource = (sqlite3 *) _pSource;
  const char * zSourceName = (const char *) _zSourceName;

  sqlite3_backup * pBackup = sqlite3_backup_init(pDest, zDestName, pSource, zSourceName);
  if (pBackup == NULL) { return NULL; }
  else {return ((Pointer) pBackup);}
}


Int32 smli_sqlite3_backup_step (Pointer _pBackup)
{
  sqlite3_backup * pBackup = (sqlite3_backup *) _pBackup;
  int r = sqlite3_backup_step (pBackup, -1);
  if (r == SQLITE_OK) {
    return 0;
  }
  else if (r == SQLITE_DONE) {
    return 1;
  }
  else {
    return 2;
  }
}

Int32 smli_sqlite3_backup_finish (Pointer _pBackup)
{
  sqlite3_backup * pBackup = (sqlite3_backup *) _pBackup;
  int r = sqlite3_backup_finish (pBackup);
  if (r == SQLITE_OK) {
    return 0;
  }
  else {
    return 1;
  }
}
