(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

structure TestBogeholz = struct

val declFileName = "bogeholz.d"

val ds = Option.valOf (DeclParser.parse_decls_file declFileName)

val _ = (Declaration.check_decls ds)	 
	
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
    
val _ = print (Declaration.print_decls ds)
	
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

val formulaFileName = "bogeholz.f"

val fmain = Option.valOf (FolParser.parse_formula_file formulaFileName)
val _ = print ((Fol.print_formula fmain) ^ "\n");
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
    
val fmain' = Tc.tc_formula ds fmain
val _ = print ((Fol.print_formula fmain') ^ "\n");	
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
    
val _ = Mc.mc_formula ds fmain'


val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

(* Open a connection to the database *)
    
structure Db = BogeholzScenario.Db;

val connection = Option.valOf (Db.newConnection BogeholzScenario.db_name)


(* Test the BogeholzScenario.resolve_eval_pred for hasattr_ext*)

val info_term = Fol.tApp ("str2infoid", [Fol.tString "_i704"])

exception Unexpected

fun sc (Solution.NOMORE) = ()
  | sc (Solution.NEXT [t1, t2]) = 
    print ((Fol.print_atom ("hasattr_ext", [t1,t2])) ^ "\n")
  | sc _ = raise Unexpected

val _ = BogeholzScenario.resolve_eval_pred  connection 
					    "hasattr_ext" 
					    [info_term, Fol.tVar "x"]
					    sc
	
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
    

(* Test the BogeholzScenario.resolve_eval_pred for inrole *)

val doctor_term = Option.valOf (FolParser.parse_term "str2prin \"_d18\"")

val role_term = Option.valOf (FolParser.parse_term "doctor (str2prin \"_r74\")")

val time_term_list
  = List.map Option.valOf 
	     [FolParser.parse_term "1984:11:23:18:59:59" (* Should fail *),
	      FolParser.parse_term "1984:11:23:19:00:00" (* Should succeed *),
	      FolParser.parse_term "1984:11:23:19:00:01" (* Should succeed *),
	      FolParser.parse_term "1984:12:01:00:00:00" (* Should succeed *),
	      FolParser.parse_term "1984:12:27:19:00:00" (* Should succeed *),
	      FolParser.parse_term "1984:12:27:19:00:01" (* Should fail *),
	      FolParser.parse_term "1985:07:01:00:00:00" (* Should fail *),
	      FolParser.parse_term "1985:09:26:19:59:59" (* Should fail *),
	      FolParser.parse_term "1985:09:26:20:00:00" (* Should succeed *)
	     ]

fun sc (Solution.NOMORE) = ()
  | sc (Solution.NEXT tl) = 
    print ((Fol.print_atom ("inrole", tl)) ^ "\n")


fun test_inrole (time_term: Fol.term) = 
    BogeholzScenario.resolve_eval_pred connection "inrole" [doctor_term, role_term, time_term] sc


val _ = List.map test_inrole time_term_list

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

(* Start testing of reduction *)

structure BR = ReductionFn (BogeholzScenario);


(* Test the resolve_guard function *)

(* We will use the guard 
  (conj
     (beginrole (str2prin "_d18") (doctor (str2prin "_r74")) x)
     (endrole (str2prin "_d18") (doctor (str2prin "_r74")) y)
     (lessthan x y)
  )

  Unfortunately, open variables will parse as term constructors, so we
  must construct this guard by hand!
*)

val guard = 
    let val pt = Option.valOf o FolParser.parse_term
    in
	Fol.gAnd [Fol.gAtom ("beginrole", [pt "str2prin \"_d18\"",
					   pt "doctor (str2prin \"_r74\")",
					   Fol.tVar "x"]),
		  Fol.gAtom ("endrole", [pt "str2prin \"_d18\"",
					 pt "doctor (str2prin \"_r74\")",
					 Fol.tVar "y"]),
		  Fol.gAtom ("lessthan", [Fol.tVar "x", Fol.tVar "y"])
		 ]
    end

val _ = print ("Guard is: " ^ (Fol.print_guard guard) ^ "\n")
val _ = print ("Solutions are: \n")

val isFinal = ref (true)
val count = ref 0

fun sc (su: Solution.subst) = 
    (List.map (fn (v, t) =>
		  print ((Fol.print_var v) ^ " = " ^ (Fol.print_term t) ^ "; ")
	      ) su;
     print "\n";  
     count := (!count) + 1
    )
    
val _ = BR.resolve_guard connection guard isFinal sc

val _ = print ("Total number of resuts: " ^ (Int.toString (!count)) ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

(* Modified query. Now the guard is:
  (conj
     (beginrole (str2prin "_d18") r x)
     (endrole (str2prin "_d18") r y)
     (lessthan x y)
 *)

val guard = 
    let val pt = Option.valOf o FolParser.parse_term
    in
	Fol.gAnd [Fol.gAtom ("beginrole", [pt "str2prin \"_d18\"",
					   Fol.tVar "r",
					   Fol.tVar "x"]),
		  Fol.gAtom ("endrole", [pt "str2prin \"_d18\"",
					 Fol.tVar "r",
					 Fol.tVar "y"]),
		  Fol.gAtom ("lessthan", [Fol.tVar "x", Fol.tVar "y"])
		 ]
    end

   
val _ = print ("Guard is: " ^ (Fol.print_guard guard) ^ "\n")
val _ = print ("Solutions are: \n")

val isFinal = ref (true)
val count = ref 0

fun sc (su: Solution.subst) = 
    (List.map (fn (v, t) =>
		  print ((Fol.print_var v) ^ " = " ^ (Fol.print_term t) ^ "; ")
	      ) su;
     print "\n";  
     count := (!count) + 1
    )
    
val _ = BR.resolve_guard connection guard isFinal sc

val _ = print ("Total number of resuts: " ^ (Int.toString (!count)) ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

(* Test reduction *)

val pf = Option.valOf o FolParser.parse_formula 

(* The formula we try is:
   (and 
     (allows (str2prin "_r0") (add (str2prin "_m0") (str2infoid "_i120")) 1969:12:31:19:00:00) 
     (mentally-incapable (str2prin "_r25") 1970:12:31:00:00:00)
     (lessthan 1969:12:31:19:00:00 1970:12:31:00:00:00)
   )

   The first and last clauses are true, so the result we expect is
     (mentally-incapable (str2prin "_r25") 1970:12:31:00:00:00)
*)

val f = pf ("(<ABC> and " ^
            "(allows (str2prin \"_r0\") (add (str2prin \"_m0\") (str2infoid \"_i120\")) 1970:12:31:19:00:00) " ^ 
            "(mentally-incapable (str2prin \"_r25\") 1971:12:31:00:00:00) " ^ 
            "(lessthan 1970:12:31:19:00:00 1971:12:31:00:00:00) " ^
            ")"
	   )

val _ = print ("Original formula is:\n" ^ (Fol.print_formula f) ^ "\n")

val f' = BR.reduce_formula connection f

val _ = print ("Reduced formula is:\n" ^ (Fol.print_formula f') ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";


(* The next formula we try is:
   (and 
     (allows (str2prin "_r0") (add (str2prin "_m0") (str2infoid "_i120")) 1969:12:31:19:00:00) 
     (lessthan 1969:12:31:19:00:00 1970:12:31:00:00:00)
   )

   Both clauses are true, so the result we expect is fTrue
*)

val f = pf ("(<ABC> and " ^
            "(allows (str2prin \"_r0\") (add (str2prin \"_m0\") (str2infoid \"_i120\")) 1970:12:31:19:00:00) " ^ 
            "(lessthan 1971:12:31:19:00:00 1972:12:31:00:00:00) " ^
	    "(true) " ^
            ")"
	   )

val _ = print ("Original formula is:\n" ^ (Fol.print_formula f) ^ "\n")

val f' = BR.reduce_formula connection f

val _ = print ("Reduced formula is:\n" ^ (Fol.print_formula f') ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

val f'' = BR.reduce_formula_no_inst connection f

val _ = print ("Reduced formula (no inst) is:\n" ^ (Fol.print_formula f'') ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

val f''' = BR.reduce_simplify  f

val _ = print ("Reduced formula (simp) is:\n" ^ (Fol.print_formula f''') ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";


(* Another test *)
(* val f = pf ("(all [k][i][u] " ^ *)
(* 	    "(and (does (add k i) u) (lessthan u 1970:02:01:00:00:00))" ^ *)
(* 	    "(or (violated (add k i) u) (lessthan u 1970:01:01:00:00:00))" ^ *)
(* 	    ")" *)
(* 	   ) *)

(* val _ = print ("Original formula is:\n" ^ (Fol.print_formula f) ^ "\n") *)

(* val f' = BR.reduce_formula connection f *)

(* val _ = print ("Reduced formula is:\n" ^ (Fol.print_formula f') ^ "\n") *)

(* val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n"; *)


val fmain_inst = BR.instantiate_formula connection fmain'

val _ = print ("Instantiated formula is:\n" ^ (Fol.print_formula fmain_inst) ^ "\n")

val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

val fmain_red = BR.reduce_formula connection fmain'

val _ = print ("Reduced formula is:\n" ^ (Fol.print_formula fmain_red) ^ "\n")
    
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";


(* Close the connection to the database *)
    
val _ = Db.closeConnection connection


end
