(*

Copyright (c) <2011>, <Deepak Garg, garg.deepak@gmail.com>
All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

structure SQLiteTest = struct

open SQLiteDb

val declFileName = "bogeholz.d"
val dbName = "bogeholz.db"

val ds = Option.valOf (DeclParser.parse_decls_file declFileName)

val _ = (Declaration.check_decls ds)	 
	
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
    
val _ = print (Declaration.print_decls ds)
	
val _ = print "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";

val db = Option.valOf (newConnection dbName)

fun columnListToSQL (cl: string list) = String.concatWith ", " cl

exception BadQuery

fun testTable (tablename: string) (columnlist: string list) (sl: Declaration.sort list) =
    let val _ = print ("Trying table: " ^ tablename ^ "\n")

	val sql = "select " ^ (columnListToSQL columnlist) ^ 
		  " from " ^ tablename
	val _ = print (sql ^ "\n")

		
	val query = case (newQuery db sql) of
			SOME q => q
		      | NONE => (print ("Bad query: (" ^ sql ^ ")\n");
				 raise BadQuery)

	fun loop () =
	    case stepQuery query of
		DATA =>
		let val stringlist = getData query
(*		    val _ = print ((String.concatWith " | " stringlist) ^ "\n")*)
		    val termlist = List.map (Option.valOf o FolParser.parse_term) stringlist
(*		    val _ = print ((Fol.print_term_list termlist) ^ "\n")*)
		    val _ = Tc.tc_term_value_list ds sl termlist
		in
		    loop ()
		end
	      | DONE => destroyQuery query
	      | ERROR => (print ("%%%%%%%%%%%%%%\nERROR in " ^ tablename ^ "\n");
			  destroyQuery query
			 )
    in
	loop()
    end

fun proc_decl (_,Declaration.declSort) = ()
  | proc_decl (_,Declaration.declConst _) = ()
  | proc_decl (_,Declaration.declPred (sml, predDB, _, SOME (tn, cl))) =
    testTable tn cl (List.map #1 sml)
  | proc_decl (_,Declaration.declPred _) = ()


val _ = (List.map proc_decl ds) handle _ => [()]

val _ = closeConnection db

end
