(*
Copyright (c) <2011>, <Limin Jia, liminjia@cmu.edu>

All rights reserved.

Use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

- Code derived from this source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Use or modification is not done for any commercial purpose or for
  monetary gain

Redistribution without permission of original copyright holder is
prohibited.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*)

(*
 ./test-hipaa testId prepDBId
 testId is the test number
 prepDBId is the file containing sql commands to set up hipaa.db properly
*)

structure TestHipaa = struct

val declFileName = "./hipaa-case-study/hipaa.d"

val ds = Option.valOf (DeclParser.parse_decls_file declFileName)
val _ = (Declaration.check_decls ds)	 
(* val _ = print "%%%%%%%%%%%%%%%%%%%%% Finished checking definitions %%%%%%%%%%%%%%%%%%\n";     *)

val formulaFileName =  "./hipaa-case-study/hipaa.f" 

val fmain = Option.valOf (FolParser.parse_formula_file formulaFileName)
(* val _ = print ((Fol.print_formula fmain) ^ "\n"); *)
(* val _ = print "%%%%%%%%%%%%%%%%%%%%%Finished parsing policies%%%%%%%%%%%%%%%%%%\n"; *)

val fmain' = Tc.tc_formula ds fmain
(* val _ = print ((Fol.print_formula fmain') ^ "\n");	*)
(* val _ = print "%%%%%%%%%%%%%%%%%%% Finished type checking %%%%%%%%%%%%%%%%%\n"; *)
    
val _ = Mc.mc_formula ds fmain' 
(* val _ = print "%%%%%%%%%%%%%%%%% Finished mode checking %%%%%%%%%%%%%%%%%%\n"; *)


(* process arguments*)

fun extractArgs () =
    case (CommandLine.arguments ())
     of [] => (NONE, NONE)
      | [id] => (SOME id, NONE)
      | id::fname::rest =>(SOME id, SOME fname)


(* Open a connection to the database *)
    
structure Db = HipaaScenario.Db;

val connection = Option.valOf (Db.newConnection HipaaScenario.db_name)

exception FailToPrep of string
exception ImpossibleDBRet
exception ErrorRunningDbQuery of string

structure HR = ReductionFn (HipaaScenario);

val timer = Timer.startRealTimer()

val fmain_red = HR.reduce_formula connection fmain' 

val timeSpent = Time.toString(Timer.checkRealTimer timer)

(* val _ = print "%%%%%%%%%%%%%%%% End Reduce %%%%%%%%%%%%%%%%%%%%%%%\n"; *)
    

val _ = 

(* update the results data base with the final running time*)
    case (#1 (extractArgs()) )of
	NONE =>( print ("%%%%%%%%%% time used: " ^ timeSpent ^"%%%%%%%%%%%%%\n");
		 print ((Fol.print_formula fmain_red) ^ "\n"))
      | SOME id => 
	let 
	    val q1 = Option.valOf(Db.newQuery connection "select (count (sender)) from disclose")
	    val nDisclosure = 
		case Db.stepQuery q1 of
		    DATA => hd(Db.getData q1)
		  | DONE => ""
		  | ERROR => (print "%%%%%%%%%%%%%%\nERROR\n"; "")
	    val q2 = Option.valOf(Db.newQuery connection "select (count (info)) from hasattrof_db")
	    val nHasattrof = 
		case Db.stepQuery q2 of
		    DATA => hd(Db.getData q2)
		  | DONE => ""
		  | ERROR => (print "%%%%%%%%%%%%%%\nERROR\n"; "")
	    val q3 = Option.valOf(Db.newQuery connection "select (count (subattr)) from attr_in_db")
	    val nAttrin = 
		case Db.stepQuery q3 of
		    DATA => hd(Db.getData q3)
		  | DONE => ""
		  | ERROR => (print "%%%%%%%%%%%%%%\nERROR\n"; "")
	in
	    (print ("UPDATE res SET totalTime = "^ timeSpent^" WHERE testId = "^id^";\n" );
	     print ("UPDATE res SET nDisclosure = "^ nDisclosure ^" WHERE testId = "^id^";\n" );
	     print ("UPDATE res SET nHasattrof = "^ nHasattrof ^" WHERE testId = "^id^";\n" );
	     print ("UPDATE res SET nAttrin = "^ nAttrin ^" WHERE testId = "^id^";\n" );
	     print ("UPDATE res SET avgTime = "
		    ^ "(SELECT totalTime from res WHERE testId ="^id
		    ^")/(SELECT nDisclosure from res WHERE testId =" ^id ^ ")" 
		    ^" WHERE testId = "^id^";\n" );
	     if  (String.size id) > 1  then 
		 let val f' = TextIO.openOut "./rform"
		 in (TextIO.output(f', (Fol.print_formula fmain_red)^"\n");
		     TextIO.closeOut f';
		     ())
		 end
	     else ()
	    )
	end

val _ = Db.closeConnection connection

end
